# -*- coding: utf-8 -*-
import numpy as np
import adaptfx as afx
nme = __name__

def min_oar_bed(keys, sets=afx.SETTING_DICT):
    # oar_minimisation is a subset of number of fraction minimisation
    # c has to be zero
    keys.c = 0
    output = min_n_frac(keys, sets=sets)

    return output


def min_n_frac(keys, sets=afx.SETTING_DICT):
    # check if keys is a dictionary from manual user
    if isinstance(keys, dict):
        keys = afx.DotDict(keys)

    if isinstance(sets, dict):
        sets = afx.DotDict(sets)

    fraction = keys.fraction
    number_of_fractions = keys.number_of_fractions
    accumulated_tumor_dose = keys.accumulated_tumor_dose
    sparing_factors_public = keys.sparing_factors_public
    prob_update = keys.prob_update
    fixed_mean = keys.fixed_mean
    fixed_std = keys.fixed_std
    shape = keys.shape
    scale = keys.scale
    shape_inv = keys.shape_inv
    scale_inv = keys.scale_inv
    tumor_goal = keys.tumor_goal
    c = keys.c
    abt = keys.abt
    abn = keys.abn
    min_dose = keys.min_dose
    max_dose = keys.max_dose
    # ---------------------------------------------------------------------- #
    # check in which fraction data should be returned for plotting
    policy_plot = 1 if sets.plot_policy == fraction else 0
    values_plot = 1 if sets.plot_values == fraction else 0
    remains_plot = 1 if sets.plot_remains == fraction else 0

    # prepare distribution
    actual_sf = sparing_factors_public[fraction]
    if prob_update == 0:
        # fixed normal distribution for sparing factor
        mean = fixed_mean
        std = fixed_std
        rv = afx.truncated_normal(mean, std, sets.sf_low, sets.sf_high)
    elif prob_update == 1:
        # update normal distribution with gamma prior
        mean = np.mean(sparing_factors_public)
        std = afx.std_posterior(sparing_factors_public, shape, scale)
        rv = afx.truncated_normal(mean, std, sets.sf_low, sets.sf_high)
    elif prob_update == 2:
        # update distribution with inverse-gamma prior
        # posterior predictive distribution is student-t distribution
        rv = afx.student_t(sparing_factors_public, shape_inv, scale_inv)
    else:
        afx.aft_error('invalid "prob_update" key was set', nme)
    # initialise distribution from random variable (rv)
    [sf, prob] = afx.sf_probdist(rv, sets.sf_low, sets.sf_high,
        sets.sf_stepsize, sets.sf_prob_threshold)
    n_sf = len(sf)

    # actionspace
    exp = afx.find_exponent(sets.dose_stepsize)
    accumulated_tumor_dose = np.round(accumulated_tumor_dose, -exp)
    remaining_bed = np.round(tumor_goal - accumulated_tumor_dose, -exp)
    n_bedsteps = int(np.round(remaining_bed / sets.dose_stepsize, 0))
    n_statesteps = int(np.round(remaining_bed / sets.state_stepsize, 0))
    max_physical_dose = afx.convert_to_physical(remaining_bed, abt)

    # automatic max_dose calculation
    if max_dose == -1:
        max_dose = max_physical_dose
    # Reduce max_dose to prohibit tumor_goal overshoot (efficiency)
    max_dose = min(max_dose, max_physical_dose)

    # actionspace in bed dose
    bedt_space = np.linspace(0, remaining_bed, n_bedsteps + 1)
    actionspace = afx.convert_to_physical(bedt_space, abt)
    range_action = (actionspace >= min_dose) & (actionspace <= max_dose)
    actionspace = actionspace[range_action]
    # bed_space to relate actionspace to oar- and tumor-dose
    bedt_space = bedt_space[range_action]
    if not range_action.any():
        # check if actionspace is empty
        bedt_space = np.array([min_dose])
        actionspace = afx.convert_to_physical(bedt_space, abt)
    bedn_space = afx.bed_calc0(actionspace, abn, actual_sf)
    n_action = len(actionspace)

    # tumor bed states for tracking dose
    tumor_limit = tumor_goal + sets.state_stepsize
    # include at least one more step for bedt
    # define number of bed_dose steps to fulfill stepsize
    bedt_states = np.linspace(accumulated_tumor_dose,
        tumor_goal, n_statesteps + 1)
    remaining_states = bedt_states - accumulated_tumor_dose
    n_bedt_states = len(bedt_states)
    
    # relate actionspace to bed and possible sparing factors
    # necessary reshape for broadcasting in value calculation
    bedn_sf_space = afx.bed_calc_matrix(actionspace, abn, sf).reshape(1, n_action, n_sf)
    # values matrix
    # dim(values) = dim(policy) = fractions_remaining * bedt * sf
    n_remaining_fractions = number_of_fractions - fraction
    values = np.zeros((n_remaining_fractions + 1, n_bedt_states, n_sf))
    if policy_plot or values_plot or remains_plot:
        policy = np.zeros((n_remaining_fractions + 1, n_bedt_states, n_sf))
        remains = np.zeros((n_remaining_fractions + 1, n_bedt_states, n_sf))
    
    finished = False
    # ------------------------------------------------------------------------------------- #
    remaining_fractions = np.arange(number_of_fractions, fraction - 1, -1)
    remaining_index = remaining_fractions - fraction
    # note that lowest fraction_state is one not zero
    # and remaining_index counts in python indices
    for fraction_index, fraction_state in zip(remaining_index, remaining_fractions):
        if remaining_bed <= 0:
            finished = True
            break

        elif fraction_state == fraction and fraction != number_of_fractions:
            # state is the actual fraction to calculate
            # e.g. in the first fraction_state there is no prior dose delivered
            # and future_bedt is equal to bedt_space
            future_values_discrete = (values[fraction_index + 1] * prob).sum(axis=1)
            future_bedt = accumulated_tumor_dose + bedt_space
            future_bedt = np.where(future_bedt > tumor_goal, tumor_limit, future_bedt)
            c_penalties = np.where(np.round(future_bedt, -exp) < tumor_goal, -c, 0)
            # for discrete matching
            # future_values = future_values_discrete[np.where(np.isclose(bedt_states, future_bedt))]
            future_values = afx.interpolate(future_bedt, bedt_states, future_values_discrete)
            vs = -bedn_space + future_values + c_penalties
            # argmax of vs along axis 0 to find best action fot the actual sf
            action_index = vs.argmax(axis=0)

            if policy_plot or values_plot or remains_plot:
                # for the policy plot
                future_bedt_full = bedt_states.reshape(n_bedt_states, 1) + bedt_space
                future_bedt_full = np.where(future_bedt_full > tumor_goal, tumor_limit, future_bedt_full)
                c_penalties_full = np.where(future_bedt_full < tumor_goal, -c, 0).reshape(n_bedt_states, n_action, 1)
                future_values_full = afx.interpolate(future_bedt_full, bedt_states, future_values_discrete)
                vs_full = -bedn_sf_space + future_values_full.reshape(n_bedt_states, n_action, 1) + c_penalties_full
                # check vs along the sf axis
                current_policy = bedt_space[vs_full.argmax(axis=1)]
                # ensure that for the goal reached the value/policy is zero (min_dose)
                current_policy[bedt_states == tumor_goal] = 0
                future_remains_discrete = (remains[fraction_index + 1] * prob).sum(axis=1)
                future_bedt_opt = current_policy + bedt_states.reshape(n_bedt_states, 1)
                future_remains = afx.interpolate(future_bedt_opt, bedt_states, future_remains_discrete)
                current_remains = np.where((current_policy - remaining_states[::-1].reshape(n_bedt_states, 1)) >= 0, 0, 1)
                # write to arrays
                policy[fraction_index] = current_policy
                values[fraction_index] = vs_full.max(axis=1)
                remains[fraction_index] = current_remains + future_remains

        elif fraction == number_of_fractions:
            # in the last fraction value is not relevant
            # max_dose is the already calculated remaining dose
            # this should compensate the discretisation of the state space
            action_index = np.abs(remaining_bed - bedt_space).argmin()
            
            if policy_plot:
                policy[fraction_index][0] = bedt_space[action_index]

        elif fraction_state == number_of_fractions:
            # final state to initialise terminal reward
            # dose remaining to be delivered, this is the actionspace in bedt
            last_bed_actions = np.round(tumor_goal - bedt_states, -exp)
            last_actions = afx.convert_to_physical(last_bed_actions, abt)
            # cut the actionspace to min and max dose constraints
            last_bed_actions = np.where(last_actions > max_dose, bedt_space[-1], last_bed_actions)
            last_bed_actions = np.where(last_actions < min_dose, bedt_space[0], last_bed_actions)
            last_actions = np.where(last_actions > max_dose, actionspace[-1], last_actions)
            last_actions = np.where(last_actions < min_dose, actionspace[0], last_actions)
            last_bedn = afx.bed_calc_matrix(last_actions, abn, sf)
            # this smooths out the penalties in underdose and overdose regions
            bedt_diff = np.round(bedt_states + last_bed_actions - tumor_goal, -exp)
            penalties = np.where(bedt_diff == 0, 0, -np.abs(bedt_diff) * sets.inf_penalty)
            # to each best action add the according penalties
            # penalties need to be reshaped for broadcasting
            vs = -last_bedn + penalties.reshape(n_bedt_states, 1)
            values[fraction_index] = vs
            # ensure that for the goal reached the value/poliy is zero (min_dose)
            values[fraction_index][bedt_states==tumor_goal] = 0

            if policy_plot:
                # policy calculation for each bedt, but sf is not considered
                policy[fraction_index] += last_bed_actions.reshape(n_bedt_states, 1)
                # ensure that for the goal reached the value/poliy is zero (min_dose)
                policy[fraction_index][bedt_states==tumor_goal] = 0

        elif fraction_state != number_of_fractions:
            # every other state but the last
            # this calculates the value function in the future fractions
            future_values_discrete = (values[fraction_index + 1] * prob).sum(axis=1)
            # bedt_states is reshaped such that numpy broadcast leads to 2D array
            future_bedt = bedt_states.reshape(n_bedt_states, 1) + bedt_space
            future_bedt = np.where(future_bedt > tumor_goal, tumor_limit, future_bedt)
            future_values = afx.interpolate(future_bedt, bedt_states, future_values_discrete)
            c_penalties = np.where(future_bedt < tumor_goal, -c, 0).reshape(n_bedt_states, n_action, 1)
            # dim(bedn_sf_space)=(1,n_action,n_sf),dim(future_values)=(n_states,n_action)
            # every row of values_penalties is transposed and copied n_sf times
            vs = -bedn_sf_space + future_values.reshape(n_bedt_states, n_action, 1) + c_penalties
            # check vs along the sf axis
            values[fraction_index] = vs.max(axis=1)
            # ensure that for the goal reached the value/poliy is zero (min_dose)
            values[fraction_index][bedt_states==tumor_goal] = 0

            if policy_plot or remains_plot:
                current_policy = bedt_space[vs.argmax(axis=1)]
                # ensure that for the goal reached the value/policy is zero (min_dose)
                current_policy[bedt_states == tumor_goal] = 0
                future_remains_discrete = (remains[fraction_index + 1] * prob).sum(axis=1)
                future_bedt_opt = current_policy + bedt_states.reshape(n_bedt_states, 1)
                future_remains = afx.interpolate(future_bedt_opt, bedt_states, future_remains_discrete)
                current_remains = np.where((current_policy - remaining_states[::-1].reshape(n_bedt_states, 1)) >= 0, 0, 1)
                # write to arrays
                policy[fraction_index] = current_policy
                remains[fraction_index] = current_remains + future_remains

    output = afx.DotDict({})

    output.physical_dose = actionspace[action_index] if not finished else np.nan
    output.tumor_dose = bedt_space[action_index] if not finished else np.nan
    output.oar_dose = bedn_space[action_index] if not finished else np.nan

    if policy_plot:
        output.policy = {}
        output.policy.val = policy
        output.policy.sf = sf
        output.policy.states = remaining_states
    if values_plot:
        output.value = {}
        output.value.val = values
        output.value.sf = sf
        output.value.states = remaining_states
    if remains_plot:
        output.remains = {}
        output.remains.val = remains
        output.remains.sf = sf
        output.remains.states = remaining_states

    return output


# -*- coding: utf-8 -*-
import numpy as np
from scipy.interpolate import interp1d, interp2d, RegularGridInterpolator
from adaptfx import std_posterior, truncated_normal, sf_probdist, bed_calc_matrix, bed_calc0, convert_to_physical, SETTING_DICT, DotDict

BED_calc_matrix = bed_calc_matrix
BED_calc0 = bed_calc0

def max_action(bed, actionspace, goal, abt=10):
    """
    Computes the maximal dose that can be delivered to the tumor
    in each fraction depending on the actual accumulated dose

    Parameters
    ----------
    bed : float
        accumulated tumor dose so far.
    actionspace : list/array
        array with all discrete dose steps.
    goal : float
        prescribed tumor dose.
    abt : float, optional
        alpha beta ratio of tumor. The default is 10.

    Returns
    -------
    sizer : integer
        gives the size of the resized actionspace to reach
        the prescribed tumor dose.

    """
    bed_actionspace = bed_calc0(actionspace, abt)
    max_action = min(max(bed_actionspace), goal - bed)
    sizer = np.argmin(np.abs(bed_actionspace - max_action))
    if bed_actionspace[sizer] < max_action:
        # make sure that with prescribing max_action
        # reached dose is not below goal
        sizer += 1

    return sizer

def argfind(bedt, value):
    """
    This function is used to find the index of certain values

    Parameters
    ----------
    bedt : list/array
        list of tumor BED in which value is searched.
    value : float
        item inside list.

    Returns
    -------
    index : integer
        index of value inside list.

    """
    index = min(range(len(bedt)), key=lambda i: abs(bedt[i] - value))
    return index


def min_oar_bed_old(keys, sets=SETTING_DICT):
    fraction = keys.fraction
    number_of_fractions = keys.number_of_fractions
    accumulated_tumor_dose = keys.accumulated_tumor_dose
    sparing_factors_public = keys.sparing_factors_public
    alpha = keys.alpha
    beta = keys.beta
    tumor_goal = keys.tumor_goal
    abt = keys.abt
    abn = keys.abn
    min_dose = keys.min_dose
    max_dose = keys.max_dose
    fixed_prob = keys.fixed_prob
    fixed_mean = keys.fixed_mean
    fixed_std = keys.fixed_std
    # ---------------------------------------------------------------------- #
    # check in which fraction policy should be returned
    policy_plot = 1 if sets.plot_policy == fraction else 0

    if fixed_prob != 1:
        mean = np.mean(
            sparing_factors_public
        )  # extract the mean and std to setup the sparingfactor distribution
        standard_deviation = std_posterior(sparing_factors_public, alpha, beta)
    if fixed_prob == 1:
        mean = fixed_mean
        standard_deviation = fixed_std
    X = truncated_normal(mean, standard_deviation, sets.sf_low, sets.sf_high)
    bedt_states = np.arange(accumulated_tumor_dose, tumor_goal, 1)
    bedt_states = np.concatenate(
        (bedt_states, [tumor_goal, tumor_goal + 1])
    )  # add an extra step outside of our prescribed tumor dose which will be penalised to make sure that we aim at the prescribe tumor dose
    [sf, prob] = sf_probdist(X, sets.sf_low, sets.sf_high,
        sets.sf_stepsize, sets.sf_prob_threshold)
    # we prepare an empty values list and open an action space which is equal to all the dose numbers that can be given in one fraction
    values = np.zeros(
        (number_of_fractions - fraction, len(bedt_states), len(sf))
    )  # 2d values list with first indice being the BED and second being the sf
    max_physical_dose = convert_to_physical(tumor_goal, abt)
    if max_dose == -1:
        max_dose = max_physical_dose
    elif max_dose > max_physical_dose:
        # if the max dose is too large we lower it, so we dont needlessly check too many actions
        max_dose = max_physical_dose
    if min_dose > max_dose:
        min_dose = max_dose - 0.1
    actionspace = np.arange(min_dose, max_dose + 0.1, 0.1)
    # now we set up the policy array which has len(BEDT)*len(sf)*len(actionspace) entries. We give each action the same probability to start with
    policy = np.zeros((number_of_fractions - fraction, len(bedt_states), len(sf)))

    for state, fraction_state in enumerate(
        np.arange(number_of_fractions, fraction - 1, -1)
    ):  # We have five fractionations with 2 special cases 0 and 4
        if (
            state == number_of_fractions - 1
        ):  # first state with no prior dose delivered so we dont loop through BEDT
            bedn = BED_calc_matrix(
                sparing_factors_public[-1], abn, actionspace
            )  # calculate all delivered doses to the Normal tissues (the penalty)
            future_values_func = interp1d(bedt_states, (values[state - 1] * prob).sum(axis=1))
            future_values = future_values_func(
                BED_calc0(actionspace, abt)
            )  # for each action and sparing factor calculate the penalty of the action and add the future value we will only have as many future values as we have actions (not sparing dependent)
            vs = -bedn + future_values
            policy4 = vs.argmax(axis=1)
        elif (
            fraction_state == fraction and fraction != number_of_fractions
        ):  # actual fraction
            actionspace_clipped = actionspace[
                0 : max_action(accumulated_tumor_dose, actionspace, tumor_goal) + 1
            ]
            bedn = BED_calc_matrix(sparing_factors_public[-1], abn, actionspace_clipped)
            future_bedt = accumulated_tumor_dose + BED_calc0(actionspace_clipped, abt)
            future_bedt[future_bedt > tumor_goal] = tumor_goal + 1
            penalties = np.zeros(future_bedt.shape)
            penalties[future_bedt > tumor_goal] = -10000
            future_values_func = interp1d(bedt_states, (values[state - 1] * prob).sum(axis=1))
            future_values = future_values_func(
                future_bedt
            )  # for each action and sparing factor calculate the penalty of the action and add the future value we will only have as many future values as we have actions (not sparing dependent)
            vs = -bedn + future_values + penalties
            policy4 = vs.argmax(axis=1)
        elif (
            fraction == number_of_fractions
        ):  # in this state no penalty has to be defined as the value is not relevant
            best_action = convert_to_physical(tumor_goal-accumulated_tumor_dose, abt)
            if accumulated_tumor_dose > tumor_goal:
                best_action = 0
            if best_action < min_dose:
                best_action = min_dose
            if best_action > max_dose:
                best_action = max_dose
            last_BEDN = BED_calc0(best_action, abn, sparing_factors_public[-1])
            policy4 = best_action * 10
        else:
            future_value_prob = (values[state - 1] * prob).sum(axis=1)
            future_values_func = interp1d(bedt_states, future_value_prob)
            for tumor_index, tumor_value in enumerate(
                bedt_states
            ):  # this and the next for loop allow us to loop through all states
                actionspace_clipped = actionspace[
                    0 : max_action(tumor_value, actionspace, tumor_goal) + 1
                ]  # we only allow the actions that do not overshoot
                bedn = BED_calc_matrix(
                    sf, abn, actionspace_clipped
                )  # this one could be done outside of the loop and only the clipping would happen inside the loop.
                bed = BED_calc_matrix(np.ones(len(sf)), abt, actionspace_clipped)
                if state != 0:
                    future_bedt = tumor_value + bed
                    future_bedt[future_bedt > tumor_goal] = tumor_goal + 1
                    future_values = future_values_func(
                        future_bedt
                    )  # for each action and sparing factor calculate the penalty of the action and add the future value we will only have as many future values as we have actions (not sparing dependent)
                    penalties = np.zeros(future_bedt.shape)
                    penalties[future_bedt > tumor_goal] = -10000
                    vs = -bedn + future_values + penalties
                    if vs.size == 0:
                        best_action = np.zeros(len(sf))
                        valer = np.zeros(len(sf))
                    else:
                        best_action = vs.argmax(axis=1)
                        valer = vs.max(axis=1)
                else:  # last state no more further values to add
                    best_action = convert_to_physical(tumor_goal-tumor_value, abt)
                    if best_action > max_dose:
                        best_action = max_dose
                    if best_action < min_dose:
                        best_action = min_dose
                    last_BEDN = BED_calc0(best_action, abn, sf)
                    future_bedt = tumor_value + BED_calc0(best_action, abt)
                    underdose_penalty = 0
                    overdose_penalty = 0
                    if future_bedt < tumor_goal:
                        underdose_penalty = (future_bedt - tumor_goal) * 10
                    if future_bedt > tumor_goal:
                        overdose_penalty = -10000
                    valer = (
                        -last_BEDN
                        + underdose_penalty * np.ones(sf.shape)
                        + overdose_penalty * np.ones(sf.shape)
                    )  # gives the value of each action for all sparing factors. elements 0-len(sparingfactors) are the Values for

                policy[state][tumor_index] = best_action
                values[state][tumor_index] = valer
    if fraction != number_of_fractions:
        optimal_action = actionspace[policy4]
    if fraction == number_of_fractions:
        optimal_action = policy4 / 10
    tumor_dose = BED_calc0(optimal_action, abt)
    oar_dose = BED_calc0(optimal_action, abn, sparing_factors_public[-1])

    output = DotDict({})

    output.physical_dose = optimal_action
    output.tumor_dose = tumor_dose
    output.oar_dose = oar_dose
    
    return output






def min_n_frac_old(keys, sets=SETTING_DICT):
    fraction = keys.fraction
    number_of_fractions = keys.number_of_fractions
    accumulated_tumor_dose = keys.accumulated_tumor_dose
    sparing_factors_public = keys.sparing_factors_public
    alpha = keys.alpha
    beta = keys.beta
    tumor_goal = keys.tumor_goal
    c = keys.c
    abt = keys.abt
    abn = keys.abn
    min_dose = keys.min_dose
    max_dose = keys.max_dose
    fixed_prob = keys.fixed_prob
    fixed_mean = keys.fixed_mean
    fixed_std = keys.fixed_std
    # ---------------------------------------------------------------------- #
    # check in which fraction policy should be returned
    policy_plot = 1 if sets.plot_policy == fraction else 0

    if fixed_prob != 1:
        mean = np.mean(
            sparing_factors_public
        )  # extract the mean and std to setup the sparingfactor distribution
        standard_deviation = std_posterior(sparing_factors_public, alpha, beta)
    if fixed_prob == 1:
        mean = fixed_mean
        standard_deviation = fixed_std
    X = truncated_normal(mean, standard_deviation, sets.sf_low, sets.sf_high)
    bedt_states = np.arange(accumulated_tumor_dose, tumor_goal, 1)
    bedt_states = np.concatenate(
        (bedt_states, [tumor_goal, tumor_goal + 1])
    )  # add an extra step outside of our prescribed tumor dose which will be penalised to make sure that we aim at the prescribe tumor dose
    [sf, prob] = sf_probdist(X, sets.sf_low, sets.sf_high,
        sets.sf_stepsize, sets.sf_prob_threshold)
    # we prepare an empty values list and open an action space which is equal to all the dose numbers that can be given in one fraction
    values = np.zeros(
        (number_of_fractions - fraction, len(bedt_states), len(sf))
    )  # 2d values list with first indice being the BED and second being the sf
    max_physical_dose = convert_to_physical(tumor_goal, abt)
    if max_dose == -1:
        max_dose = max_physical_dose
    elif max_dose > max_physical_dose:
        # if the max dose is too large we lower it, so we dont needlessly check too many actions
        max_dose = max_physical_dose
    if min_dose > max_dose:
        min_dose = max_dose - 0.1
    actionspace = np.arange(min_dose, max_dose + 0.1, 0.1)
    # now we set up the policy array which has len(BEDT)*len(sf)*len(actionspace) entries. We give each action the same probability to start with
    policy = np.zeros((number_of_fractions - fraction, len(bedt_states), len(sf)))

    for state, fraction_state in enumerate(
        np.arange(number_of_fractions, fraction -1, -1)
    ):
        if (
            state == number_of_fractions - 1 #fraction_state == 1
        ):  # first state with no prior dose delivered so we dont loop through BEDT
            bedn = bed_calc_matrix(
                sparing_factors_public[-1], abn, actionspace
            )  # calculate all delivered doses to the Normal tissues (the penalty)
            future_bedt = bed_calc0(actionspace, abt)
            future_values_func = interp1d(bedt_states, (values[state - 1] * prob).sum(axis=1))
            future_values = future_values_func(
                future_bedt
            )  # for each action and sparing factor calculate the penalty of the action and add the future value we will only have as many future values as we have actions (not sparing dependent)
            c_penalties = np.zeros(future_bedt.shape)
            c_penalties[future_bedt < tumor_goal] = -c
            vs = -bedn + c_penalties + future_values
            policy4 = vs.argmax(axis=1)
        elif (
            accumulated_tumor_dose >= tumor_goal
        ):
            best_action = 0
            last_BEDN = bed_calc0(best_action, abn, sparing_factors_public[-1])
            policy4 = 0
            break
        elif (
            fraction_state == fraction and fraction != number_of_fractions
        ):  # actual fraction is first state to calculate
            actionspace_clipped = actionspace[
                0 : max_action(accumulated_tumor_dose, actionspace, tumor_goal) + 1
            ]
            bedn = bed_calc_matrix(sparing_factors_public[-1], abn, actionspace_clipped)
            future_bedt = accumulated_tumor_dose + bed_calc0(actionspace_clipped, abt)
            future_bedt[future_bedt > tumor_goal] = tumor_goal + 1
            penalties = np.zeros(future_bedt.shape)
            c_penalties = np.zeros(future_bedt.shape)
            penalties[future_bedt > tumor_goal] = 0 #overdosing is indirectly penalised with BEDN
            c_penalties[future_bedt < tumor_goal] = -c
            future_values_func = interp1d(bedt_states, (values[state - 1] * prob).sum(axis=1))
            future_values = future_values_func(
                future_bedt
            )  # for each action and sparing factor calculate the penalty of the action and add the future value we will only have as many future values as we have actions (not sparing dependent)
            vs = -bedn + c_penalties + future_values + penalties
            policy4 = vs.argmax(axis=1)
        elif (
            fraction == number_of_fractions
        ):  # in this state no penalty has to be defined as the value is not relevant
            best_action = convert_to_physical(tumor_goal-accumulated_tumor_dose, abt)
            if accumulated_tumor_dose > tumor_goal:
                best_action = 0
            if best_action < min_dose:
                best_action = min_dose
            if best_action > max_dose:
                best_action = max_dose
            last_BEDN = bed_calc0(best_action, abn, sparing_factors_public[-1])
            policy4 = best_action * 10
        else:
            future_value_prob = (values[state - 1] * prob).sum(axis=1)
            future_values_func = interp1d(bedt_states, future_value_prob)
            for tumor_index, tumor_value in enumerate(
                bedt_states
            ):  # this and the next for loop allow us to loop through all states
                actionspace_clipped = actionspace[
                    0 : max_action(tumor_value, actionspace, tumor_goal) + 1
                ]  # we only allow the actions that do not overshoot
                bedn = bed_calc_matrix(
                    sf, abn, actionspace_clipped
                )  # this one could be done outside of the loop and only the clipping would happen inside the loop.
                bed = bed_calc_matrix(np.ones(len(sf)), abt, actionspace_clipped)
                if state != 0 and tumor_value < tumor_goal:
                    future_bedt = tumor_value + bed
                    future_bedt[future_bedt > tumor_goal] = tumor_goal + 1
                    future_values = future_values_func(
                        future_bedt
                    )  # for each action and sparing factor calculate the penalty of the action and add the future value we will only have as many future values as we have actions (not sparing dependent)
                    c_penalties = np.zeros(future_bedt.shape)
                    c_penalties[future_bedt < tumor_goal] = -c
                    vs = -bedn + c_penalties + future_values
                    if vs.size == 0:
                        best_action = np.zeros(len(sf))
                        valer = np.zeros(len(sf))
                    else:
                        best_action = actionspace_clipped[vs.argmax(axis=1)]
                        valer = vs.max(axis=1)
                elif tumor_value > tumor_goal: #calculate value for terminal case
                    best_action = 0
                    future_bedt = tumor_value
                    valer = (
                        + np.zeros(sf.shape)
                    )  
                else:  # last state no more further values to add
                    best_action = convert_to_physical(tumor_goal-tumor_value, abt)
                    if best_action > max_dose:
                        best_action = max_dose
                    if best_action < min_dose:
                        best_action = min_dose
                    last_BEDN = bed_calc0(best_action, abn, sf)
                    future_bedt = tumor_value + bed_calc0(best_action, abt)
                    underdose_penalty = 0
                    overdose_penalty = 0
                    if future_bedt < tumor_goal:
                        underdose_penalty = -10000
                    if future_bedt > tumor_goal:
                        overdose_penalty = 0
                    valer = (
                        -last_BEDN
                        + underdose_penalty * np.ones(sf.shape)
                        + overdose_penalty * np.ones(sf.shape)
                    )  # gives the value of each action for all sparing factors. elements 0-len(sparingfactors) are the Values for

                policy[state][tumor_index] = best_action
                values[state][tumor_index] = valer

    if sets.plot_policy or sets.plot_value:
        policy = np.concatenate([np.zeros((1, len(bedt_states), len(sf))), policy[::-1]])
        values = np.concatenate([np.zeros((1, len(bedt_states), len(sf))), values[::-1]])
    if fraction != number_of_fractions:
        optimal_action = actionspace[policy4]
    if fraction == number_of_fractions:
        optimal_action = policy4 / 10
    tumor_dose = bed_calc0(optimal_action, abt)
    oar_dose = bed_calc0(optimal_action, abn, sparing_factors_public[-1])

    output = DotDict({})

    output.physical_dose = optimal_action
    output.tumor_dose = tumor_dose
    output.oar_dose = oar_dose
    
    return output







def max_tumor_bed_old(keys, sets=SETTING_DICT):
    fraction = keys.fraction
    number_of_fractions = keys.number_of_fractions
    accumulated_oar_dose = keys.accumulated_tumor_dose
    sparing_factors_public = keys.sparing_factors_public
    alpha = keys.alpha
    beta = keys.beta
    oar_limit = keys.oar_limit
    abt = keys.abt
    abn = keys.abn
    min_dose = keys.min_dose
    max_dose = keys.max_dose
    fixed_prob = keys.fixed_prob
    fixed_mean = keys.fixed_mean
    fixed_std = keys.fixed_std
    # ---------------------------------------------------------------------- #
    # check in which fraction policy should be returned
    policy_plot = 1 if sets.plot_policy == fraction else 0
    
    actual_sparing = sparing_factors_public[-1]
    if fixed_prob != 1:
        mean = np.mean(
            sparing_factors_public
        )  # extract the mean and std to setup the sparingfactor distribution
        standard_deviation = std_posterior(sparing_factors_public, alpha, beta)
    if fixed_prob == 1:
        mean = fixed_mean
        standard_deviation = fixed_std
    X = truncated_normal(mean, standard_deviation, sets.sf_low, sets.sf_high)
    [sf, prob] = sf_probdist(X, sets.sf_low, sets.sf_high,
        sets.sf_stepsize, sets.sf_prob_threshold)

    bedn = np.arange(accumulated_oar_dose, oar_limit + 1.6, 1)
    values = np.zeros(
        ((number_of_fractions - fraction), len(bedn), len(sf))
    )  # 2d values list with first indice being the BED and second being the sf
    max_physical_dose = convert_to_physical(oar_limit, abn, actual_sparing)
    if max_dose == -1:
        max_dose = max_physical_dose
    elif max_dose > max_physical_dose:
        # if the max dose is too large we lower it, so we dont needlessly check too many actions
        max_dose = max_physical_dose
    if min_dose > max_dose:
        min_dose = max_dose - 0.1
    actionspace = np.arange(min_dose, max_dose + 0.1, 0.1)
    policy = np.zeros(((number_of_fractions - fraction), len(bedn), len(sf)))
    upperbound = oar_limit + 1

    delivered_doses = BED_calc_matrix(sf, abn, actionspace)
    bedt_reward = BED_calc_matrix(
        1, abt, actionspace
    )  # this is the reward for the dose deposited inside the tumor.
    bedt_transformed, meaningless = np.meshgrid(bedt_reward, np.zeros(len(sf)))

    for index, frac_state in enumerate(
        np.arange(fraction, number_of_fractions + 1)
    ):  # We have number_of fraction fractionations with 2 special cases 0 and number_of_fractions-1 (last first fraction)
        if (
            index == number_of_fractions - 1
        ):  # first state with no prior dose delivered so we dont loop through BEDT
            future_bed = accumulated_oar_dose + delivered_doses
            future_bed[
                future_bed > oar_limit
            ] = upperbound  # any dose surpassing the upper bound will be set to the upper bound which will be penalised strongly
            value_interpolation = interp2d(sf, bedn, values[index - 1])
            future_value = np.zeros(len(sf) * len(actionspace) * len(sf)).reshape(
                len(sf), len(actionspace), len(sf)
            )
            for actual_sf in range(0, len(sf)):
                future_value[actual_sf] = value_interpolation(sf, future_bed[actual_sf])
            future_values_prob = (future_value * prob).sum(
                axis=2
            )  # in this array are all future values multiplied with the probability of getting there. shape = sparing factors x actionspace
            penalties = np.zeros(future_bed.shape)
            penalties[
                future_bed > oar_limit
            ] = (
                -1000
            )  # penalising in each fraction is needed. If not, once the algorithm reached the upper bound, it would just deliver maximum dose over and over again
            vs = future_values_prob + bedt_transformed + penalties

            actual_policy = vs.argmax(axis=1)
            actual_value = vs.max(axis=1)

        else:
            if (
                index == number_of_fractions - fraction
            ):  # if we are in the actual fraction we do not need to check all possible BED states but only the one we are in
                if fraction != number_of_fractions:
                    future_bed = accumulated_oar_dose + delivered_doses
                    overdosing = (future_bed - oar_limit).clip(min=0)
                    penalties_overdose = (
                        overdosing * -1000
                    )  # additional penalty when overdosing is needed when choosing a minimum dose to be delivered
                    future_bed[
                        future_bed > oar_limit
                    ] = upperbound  # any dose surpassing the upper bound will be set to the upper bound which will be penalised strongly
                    value_interpolation = interp2d(
                        sf, bedn, values[index - 1]
                    )
                    future_value = np.zeros(
                        len(sf) * len(actionspace) * len(sf)
                    ).reshape(len(sf), len(actionspace), len(sf))
                    for actual_sf in range(0, len(sf)):
                        future_value[actual_sf] = value_interpolation(
                            sf, future_bed[actual_sf]
                        )
                    future_values_prob = (future_value * prob).sum(
                        axis=2
                    )  # in this array are all future values multiplied with the probability of getting there. shape = sparing factors x actionspace
                    penalties = np.zeros(future_bed.shape)
                    penalties[
                        future_bed > oar_limit
                    ] = (
                        -1000
                    )  # penalising in each fraction is needed. If not, once the algorithm reached the upper bound, it would just deliver maximum dose over and over again
                    vs = (
                        future_values_prob
                        + bedt_transformed
                        + penalties
                        + penalties_overdose
                    )
                    actual_policy = vs.argmax(axis=1)
                    actual_value = vs.max(axis=1)
                else:
                    best_action = convert_to_physical(oar_limit-accumulated_oar_dose, abn, sf)
                    if accumulated_oar_dose > oar_limit:
                        best_action = np.ones(best_action.shape) * min_dose
                    best_action[best_action < min_dose] = min_dose
                    best_action[best_action > max_dose] = max_dose
                    actual_policy = best_action * 10
                    actual_value = BED_calc0(
                        best_action, abt
                    )  # we do not need to penalise, as this value is not relevant.
            else:
                for bed_index, bed_value in enumerate(
                    bedn
                ):  # this and the next for loop allow us to loop through all states
                    future_bed = delivered_doses + bed_value
                    overdosing = (future_bed - oar_limit).clip(min=0)
                    future_bed[
                        future_bed > oar_limit
                    ] = upperbound  # any dose surpassing 90.1 is set to 90.1
                    if index == 0:  # last state no more further values to add
                        best_action = convert_to_physical(oar_limit-bed_value, abn, sf)
                        best_action[best_action < min_dose] = min_dose
                        best_action[best_action > max_dose] = max_dose
                        future_bed = BED_calc0(sf, abn, best_action) + bed_value
                        overdosing = (future_bed - oar_limit).clip(min=0)
                        penalties_overdose = (
                            overdosing * -1000
                        )  # additional penalty when overdosing is needed when choosing a minimum dose to be delivered
                        future_bed[
                            future_bed > oar_limit + 0.0001
                        ] = upperbound  # 0.0001 is added due to some rounding problems
                        penalties = np.zeros(future_bed.shape)
                        if bed_value < oar_limit:
                            penalties[future_bed == upperbound] = -1000
                        values[index][bed_index] = (
                            BED_calc0(best_action, abt) + penalties + penalties_overdose
                        )
                        policy[index][bed_index] = (best_action - min_dose) * 10
                    else:
                        penalties = np.zeros(future_bed.shape)
                        penalties[future_bed == upperbound] = -1000
                        penalties_overdose = (
                            overdosing * -1000
                        )  # additional penalty when overdosing is needed when choosing a minimum dose to be delivered
                        value_interpolation = interp2d(
                            sf, bedn, values[index - 1]
                        )
                        future_value = np.zeros((len(sf), len(actionspace), len(sf)))
                        for actual_sf in range(0, len(sf)):
                            future_value[actual_sf] = value_interpolation(
                                sf, future_bed[actual_sf]
                            )
                        future_values_prob = (future_value * prob).sum(axis=2)
                        vs = (
                            future_values_prob
                            + bedt_transformed
                            + penalties
                            + penalties_overdose
                        )
                        best_action = vs.argmax(axis=1)
                        valer = vs.max(axis=1)
                        policy[index][bed_index] = best_action
                        values[index][bed_index] = valer
    index_sf = argfind(sf, actual_sparing)
    if fraction != number_of_fractions:
        dose_delivered_tumor = BED_calc0(actionspace[actual_policy[index_sf]], abt)
        dose_delivered_oar = BED_calc0(
            actionspace[actual_policy[index_sf]], abn, actual_sparing
        )
        total_dose_delivered_oar = dose_delivered_oar + accumulated_oar_dose
        actual_dose_delivered = actionspace[actual_policy[index_sf]]
    else:
        dose_delivered_tumor = BED_calc0(actual_policy[index_sf] / 10, abt)
        dose_delivered_oar = BED_calc0(
            actual_policy[index_sf] / 10, abn, actual_sparing
        )
        total_dose_delivered_oar = dose_delivered_oar + accumulated_oar_dose
        actual_dose_delivered = actual_policy[index_sf] / 10

    output = DotDict({})

    output.physical_dose = actual_dose_delivered
    output.tumor_dose = dose_delivered_tumor
    output.oar_dose = dose_delivered_oar
    
    return output






def min_oar_max_tumor_old(keys, sets=SETTING_DICT):
    fraction = keys.fraction
    number_of_fractions = keys.number_of_fractions
    accumulated_tumor_dose = keys.accumulated_tumor_dose
    accumulated_oar_dose = keys.accumulated_oar_dose
    sparing_factors_public = keys.sparing_factors_public
    alpha = keys.alpha
    beta = keys.beta
    tumor_goal = keys.tumor_goal
    oar_limit = keys.oar_limit
    abt = keys.abt
    abn = keys.abn
    min_dose = keys.min_dose
    max_dose = keys.max_dose
    fixed_prob = keys.fixed_prob
    fixed_mean = keys.fixed_mean
    fixed_std = keys.fixed_std
    # ---------------------------------------------------------------------- #
    # check in which fraction policy should be returned
    policy_plot = 1 if sets.plot_policy == fraction else 0

    if fixed_prob != 1:
        mean = np.mean(
            sparing_factors_public
        )  # extract the mean and std to setup the sparingfactor distribution
        standard_deviation = std_posterior(sparing_factors_public, alpha, beta)
    if fixed_prob == 1:
        mean = fixed_mean
        standard_deviation = fixed_std
    X = truncated_normal(mean, standard_deviation, sets.sf_low, sets.sf_high)
    [sf, prob] = sf_probdist(X, sets.sf_low, sets.sf_high,
        sets.sf_stepsize, sets.sf_prob_threshold)
    underdosepenalty = 10
    print(accumulated_oar_dose, accumulated_tumor_dose)
    bedt = np.arange(accumulated_tumor_dose, tumor_goal, 1)  # tumordose
    bedn = np.arange(accumulated_oar_dose, oar_limit, 1)  # OAR dose
    bedn = np.concatenate((bedn, [oar_limit, oar_limit + 1]))
    bedt = np.concatenate((bedt, [tumor_goal, tumor_goal + 1]))
    values = np.zeros(
        [(number_of_fractions - fraction), len(bedt), len(bedn), len(sf)]
    )  # 2d values list with first indice being the BED and second being the sf
    max_physical_dose = convert_to_physical(tumor_goal, abt)
    if max_dose == -1:
        max_dose = max_physical_dose
    elif max_dose > max_physical_dose:
        # if the max dose is too large we lower it, so we dont needlessly check too many actions
        max_dose = max_physical_dose
    if min_dose > max_dose:
        min_dose = max_dose - 0.1
    actionspace = np.arange(min_dose, max_dose + 0.1, 0.1)
    policy = np.zeros(
        ((number_of_fractions - fraction), len(bedt), len(bedn), len(sf))
    )
    upperbound_normal_tissue = oar_limit + 1
    upperbound_tumor = tumor_goal + 1

    oar_dose = BED_calc_matrix(
        sf, abn, actionspace
    )  # calculates the dose that is deposited into the normal tissue for all sparing factors
    tumor_dose = BED_calc_matrix(1, abt, actionspace)[
        0
    ]  # this is the dose delivered to the tumor
    actual_fraction_sf = argfind(sf, sparing_factors_public[-1])

    for index, frac_state_plus in enumerate(
        np.arange(number_of_fractions + 1, fraction, -1)
    ):  # We have five fractionations with 2 special cases 0 and 4
        frac_state = frac_state_plus - 1
        if (
            frac_state == 1
        ):  # first state with no prior dose delivered so we dont loop through BEDNT
            future_oar = accumulated_oar_dose + oar_dose[actual_fraction_sf]
            future_tumor = accumulated_tumor_dose + tumor_dose
            future_oar[
                future_oar > oar_limit
            ] = upperbound_normal_tissue  # any dose surpassing the upper bound will be set to the upper bound which will be penalised strongly
            future_tumor[future_tumor > tumor_goal] = upperbound_tumor
            future_values_prob = (values[index - 1] * prob).sum(
                axis=2
            )  # future values of tumor and oar state
            value_interpolation = RegularGridInterpolator(
                (bedt, bedn), future_values_prob
            )
            future_value_actual = value_interpolation(
                np.array([future_tumor, future_oar]).T
            )
            vs = future_value_actual - oar_dose[actual_fraction_sf]
            actual_policy = vs.argmax(axis=0)

        elif (
            frac_state == fraction
        ):  # if we are in the actual fraction we do not need to check all possible BED states but only the one we are in
            if fraction != number_of_fractions:
                future_oar = accumulated_oar_dose + oar_dose[actual_fraction_sf]
                overdosing = (future_oar - oar_limit).clip(min=0)
                future_oar[
                    future_oar > oar_limit
                ] = upperbound_normal_tissue  # any dose surpassing the upper bound will be set to the upper bound which will be penalised strongly
                future_tumor = accumulated_tumor_dose + tumor_dose
                future_tumor[future_tumor > tumor_goal] = upperbound_tumor
                future_values_prob = (values[index - 1] * prob).sum(
                    axis=2
                )  # future values of tumor and oar state
                value_interpolation = RegularGridInterpolator(
                    (bedt, bedn), future_values_prob
                )
                penalties = (
                    overdosing * -10000000000
                )  # additional penalty when overdosing is needed when choosing a minimum dose to be delivered
                future_value_actual = value_interpolation(
                    np.array([future_tumor, future_oar]).T
                )
                vs = future_value_actual - oar_dose[actual_fraction_sf] + penalties
                actual_policy = vs.argmax(axis=0)
            else:
                sf_end = sparing_factors_public[-1]
                best_action_oar = convert_to_physical(oar_limit-accumulated_oar_dose, abn, sf_end)
                best_action_tumor = convert_to_physical(tumor_goal-accumulated_tumor_dose, abt)
                best_action = np.min([best_action_oar, best_action_tumor], axis=0)
                if accumulated_oar_dose > oar_limit or accumulated_tumor_dose > tumor_goal:
                    best_action = np.ones(best_action.shape) * min_dose
                if best_action > max_dose:
                    best_action = max_dose
                if best_action < min_dose:
                    best_action = min_dose

                future_tumor = accumulated_tumor_dose + BED_calc0(best_action, abt)
                future_oar = accumulated_oar_dose + BED_calc0(best_action, abn, sf_end)
                actual_policy = best_action * 10
        else:
            future_values_prob = (values[index - 1] * prob).sum(
                axis=2
            )  # future values of tumor and oar state
            value_interpolation = RegularGridInterpolator(
                (bedt, bedn), future_values_prob
            )  # interpolation function
            for tumor_index, tumor_value in enumerate(bedt):
                for oar_index, oar_value in enumerate(
                    bedn
                ):  # this and the next for loop allow us to loop through all states
                    future_oar = oar_dose + oar_value
                    overdosing = (future_oar - oar_limit).clip(min=0)
                    future_oar[
                        future_oar > oar_limit
                    ] = upperbound_normal_tissue  # any dose surpassing 90.1 is set to 90.1
                    future_tumor = tumor_value + tumor_dose
                    future_tumor[
                        future_tumor > tumor_goal
                    ] = upperbound_tumor  # any dose surpassing the tumor bound is set to tumor_bound + 0.1

                    if (
                        frac_state == number_of_fractions
                    ):  # last state no more further values to add
                        best_action_oar = convert_to_physical(oar_limit-oar_value, abn, sf) 
                        # calculate maximal dose that can be delivered to OAR and tumor
                        best_action_tumor = convert_to_physical(tumor_goal-tumor_value, abt, sf*0+1)
                        best_action = np.min(
                            [best_action_oar, best_action_tumor], axis=0
                        )  # take the smaller of both doses to not surpass the limit
                        best_action[best_action > max_dose] = max_dose
                        best_action[best_action < min_dose] = min_dose
                        if (
                            oar_value > oar_limit or tumor_value > tumor_goal
                        ):  # if the limit is already surpassed we add a penaltsy
                            best_action = np.ones(best_action.shape) * min_dose
                        future_oar = oar_value + BED_calc0(best_action, abn, sf)
                        future_tumor = tumor_value + BED_calc0(best_action, abt, 1)
                        overdose_penalty2 = np.zeros(
                            best_action.shape
                        )  # we need a second penalty if we overdose in the last fraction
                        overdose_penalty3 = np.zeros(best_action.shape)
                        overdose_penalty2[
                            future_tumor > tumor_goal + 0.0001
                        ] = -100000000000
                        overdose_penalty3[
                            future_oar > oar_limit + 0.0001
                        ] = (
                            -100000000000
                        )  # A small number has to be added as sometimes 90. > 90 was True
                        end_penalty = (
                            -abs(future_tumor - tumor_goal) * underdosepenalty
                        )  # the farther we are away from the prescribed dose, the higher the penalty. Under- and overdosing is punished
                        end_penalty_oar = (
                            -(future_oar - oar_limit).clip(min=0) * 1000
                        )  # if overdosing the OAR is not preventable, the overdosing should stay as low as possible
                        values[index][tumor_index][oar_index] = (
                            end_penalty
                            - BED_calc0(best_action, abn, sf)
                            + overdose_penalty2
                            + overdose_penalty3
                            + end_penalty_oar
                        )  # we also substract all the dose delivered to the OAR so the algorithm tries to minimise it
                        policy[index][tumor_index][oar_index] = best_action * 10
                    else:
                        future_value = np.zeros([len(sf), len(actionspace)])
                        for actual_sf in range(0, len(sf)):
                            future_value[actual_sf] = value_interpolation(
                                np.array([future_tumor, future_oar[actual_sf]]).T
                            )
                        penalties = (
                            overdosing * -10000000000
                        )  # additional penalty when overdosing is needed when choosing a minimum dose to be delivered
                        vs = future_value - oar_dose + penalties
                        best_action = vs.argmax(axis=1)
                        valer = vs.max(axis=1)
                        policy[index][tumor_index][oar_index] = best_action
                        values[index][tumor_index][oar_index] = valer
    if fraction != number_of_fractions:
        physical_dose = actionspace[actual_policy]
    else:
        physical_dose = actual_policy / 10
    tumor_dose = BED_calc0(physical_dose, abt)
    oar_dose = BED_calc0(physical_dose, abn, sparing_factors_public[-1])

    output = DotDict({})

    output.physical_dose = physical_dose
    output.tumor_dose = tumor_dose
    output.oar_dose = oar_dose
    
    return output


# -*- coding: utf-8 -*-
"""
interpolation program for OAR minimisation whole plan with minimum and maximum physical dose (to tumor)
2D interpolation. In this program the optimal fraction doses are compueted based on a prescribed tumor dose while minimizing OAR BED.
single_fraction allows to compute single fraction doses, while whole_plan computes the doses for a whole treatment plan (when all sparing factors are known).
whole_plan_print prints the doses in a well-aranged manner.
"""

import numpy as np
from scipy.interpolate import interp1d
from scipy.stats import gamma, truncnorm


def data_fit(data):
    """
    This function fits the alpha and beta value for the prior

    Parameters
    ----------
    data : array
        a nxk matrix with n the amount of patints and k the amount of sparing factors per patient.

    Returns
    -------
    list
        alpha and beta hyperparameter.
    """
    standard_devs = data.std(axis=1)
    alpha, loc, beta = gamma.fit(standard_devs, floc=0)
    return [alpha, beta]


def get_truncated_normal(mean=0, sd=1, low=0.01, upp=10):
    """
    produces a truncated normal distribution

    Parameters
    ----------
    mean : float, optional
        The default is 0.
    sd : float, optional
        The default is 1.
    low : float, optional
        The default is 0.01.
    upp : float, optional
        The default is 10.

    Returns
    -------
    scipy.stats._distn_infrastructure.rv_frozen
        distribution function.

    """
    return truncnorm((low - mean) / sd, (upp - mean) / sd, loc=mean, scale=sd)


def probdist(X):
    """
    This function produces a probability distribution based on the normal distribution X

    Parameters
    ----------
    X : scipy.stats._distn_infrastructure.rv_frozen
        distribution function.

    Returns
    -------
    prob : list
        list with probabilities for each sparing factor.

    """
    prob = np.zeros(170)
    idx = 0
    for i in np.arange(0.01, 1.71, 0.01):
        prob[idx] = X.cdf(i + 0.004999999999999999999) - X.cdf(i - 0.005)
        idx += 1
    return prob


def std_calc(measured_data, alpha, beta):
    """
    calculates the most likely standard deviation for a list of k sparing factors and a gamma conjugate prior
    measured_data: list/array with k sparing factors

    Parameters
    ----------
    measured_data : list/array
        list/array with k sparing factors
    alpha : float
        shape of gamma distribution
    beta : float
        scale of gamma distrinbution

    Returns
    -------
    std : float
        most likely std based on the measured data and gamma prior

    """
    n = len(measured_data)
    std_values = np.arange(0.00001, 0.5, 0.00001)
    likelihood_values = np.zeros(len(std_values))
    for index, value in enumerate(std_values):
        likelihood_values[index] = (
            value ** (alpha - 1)
            / value ** (n - 1)
            * np.exp(-1 / beta * value)
            * np.exp(-np.var(measured_data) / (2 * (value**2 / n)))
        )  # here i have to check whether it is good.
    std = std_values[np.argmax(likelihood_values)]
    return std


def BED_calc0(dose, ab, sparing=1):
    """
    calculates the BED for a specific dose

    Parameters
    ----------
    dose : float
        physical dose to be delivered.
    ab : float
        alpha-beta ratio of tissue.
    sparing : float, optional
        sparing factor. The default is 1 (tumor).

    Returns
    -------
    BED : float
        BED to be delivered based on dose, sparing factor and alpha-beta ratio.

    """
    BED = sparing * dose * (1 + (sparing * dose) / ab)
    return BED


def BED_calc_matrix(sf, ab, actionspace):
    """
    calculates the BED for an array of values

    Parameters
    ----------
    sf : list/array
        list of sparing factors to calculate the correspondent BED.
    ab : float
        alpha-beta ratio of tissue.
    actionspace : list/array
        doses to be delivered.

    Returns
    -------
    BED : List/array
        list of all future BEDs based on the delivered doses and sparing factors.

    """
    BED = np.outer(sf, actionspace) * (
        1 + np.outer(sf, actionspace) / ab
    )  # produces a sparing factors x actions space array
    return BED


def max_action(bed, actionspace, goal, abt=10):
    """
    Computes the maximal dose that can be delivered to the tumor in each fraction depending on the actual accumulated dose

    Parameters
    ----------
    bed : float
        accumulated tumor dose so far.
    actionspace : list/array
        array with all discrete dose steps.
    goal : float
        prescribed tumor dose.
    abt : float, optional
        alpha beta ratio of tumor. The default is 10.

    Returns
    -------
    sizer : integer
        gives the size of the resized actionspace to reach the prescribed tumor dose.

    """
    max_action = min(max(BED_calc0(actionspace, abt)), goal - bed)
    sizer = np.argmin(np.abs(BED_calc0(actionspace, abt) - max_action))

    return sizer


def argfind(BEDT, value):
    """
    This function is used to find the index of certain values.
    searched_list: list/array with values
    value: value that should be inside the list
    return: index of value

    Parameters
    ----------
    searched_list : list/array
        list in which our searched value is.
    value : float
        item inside list.

    Returns
    -------
    index : integer
        index of value inside list.

    """
    index = min(range(len(BEDT)), key=lambda i: abs(BEDT[i] - value))
    return index


def value_eval(
    fraction,
    number_of_fractions,
    accumulated_dose,
    sparing_factors,
    alpha,
    beta,
    goal,
    abt,
    abn,
    min_dose=0,
    max_dose=22.3,
    fixed_prob=0,
    fixed_mean=0,
    fixed_std=0,
):
    """
    calculates the optimal dose for the desired fraction.

    Parameters
    ----------
    fraction : integer
        Number of the actual fraction.
    number_of_fractions : integer
        number of fractions that will be delivered.
    accumulated_dose : float
        accumulated tumor BED.
    sparing_factors : list/array
        list or array of all observed sparing factors. Include planning session sparing factor!.
    alpha : float
        alpha hyperparameter of std prior derived from previous patients.
    beta : float
        beta hyperparameter of std prior derived from previous patients
    goal : float
        prescribed tumor BED.
    abt : float, optional
        alpha beta ratio of the tumor. The default is 10.
    abn : float, optional
        alpha beta ratio of the organ at risk. The default is 3.
    min_dose : float
        minimal physical doses to be delivered in one fraction. The doses are aimed at PTV 95
    max_dose : float
        maximal physical doses to be delivered in one fraction. The doses are aimed at PTV 95
    fixed_prob : int
        this variable is to turn on a fixed probability distribution. If the variable is not used (0), then the probability will be updated. If the variable is turned to 1, the inserted mean and std will be used for a fixed sparing factor distribution
    fixed_mean: float
        mean of the fixed sparing factor normal distribution
    std_fixed: float
        standard deviation of the fixed sparing factor normal distribution

    Returns
    -------
    list
        Returns list with policies,relevant sparing factors range, physical dose to be delivered, tumor BED to be delivered, OAR BED to be delivered.

    """
    if fixed_prob != 1:
        mean = np.mean(
            sparing_factors
        )  # extract the mean and std to setup the sparingfactor distribution
        standard_deviation = std_calc(sparing_factors, alpha, beta)
    if fixed_prob == 1:
        mean = fixed_mean
        standard_deviation = fixed_std
    X = get_truncated_normal(mean=mean, sd=standard_deviation, low=0.01, upp=1.3)
    BEDT = np.arange(accumulated_dose, goal, 1)
    BEDT = np.concatenate(
        (BEDT, [goal, goal + 1])
    )  # add an extra step outside of our prescribed tumor dose which will be penalized to make sure that we aim at the prescribe tumor dose
    prob = np.array(probdist(X))
    sf = np.arange(0.01, 1.71, 0.01)
    sf = sf[prob > 0.00001]
    prob = prob[prob > 0.00001]
    # we prepare an empty values list and open an action space which is equal to all the dose numbers that can be given in one fraction
    Values = np.zeros(
        (number_of_fractions - fraction, len(BEDT), len(sf))
    )  # 2d values list with first indice being the BED and second being the sf
    if max_dose > (-1 + np.sqrt(1**2 + 4 * 1**2 * (goal) / abt)) / (
        2 * 1**2 / abt
    ):  # if the max dose is too large we lower it, so we dont needlessly check too many actions
        max_dose = np.round(
            (-1 + np.sqrt(1**2 + 4 * 1**2 * (goal) / abt)) / (2 * 1**2 / abt), 1
        )
    if min_dose > max_dose:
        min_dose = max_dose - 0.1
    actionspace = np.arange(min_dose, max_dose + 0.01, 0.1)
    # now we set up the policy array which has len(BEDT)*len(sf)*len(actionspace) entries. We give each action the same probability to start with
    policy = np.zeros((number_of_fractions - fraction, len(BEDT), len(sf)))

    for state, fraction_state in enumerate(
        np.arange(number_of_fractions + 1, fraction, -1)
    ):  # We have five fractionations with 2 special cases 0 and 4
        fraction_state = fraction_state - 1
        if (
            state == number_of_fractions - 1
        ):  # first state with no prior dose delivered so we dont loop through BEDT
            BEDN = BED_calc_matrix(
                sparing_factors[-1], abn, actionspace
            )  # calculate all delivered doses to the Normal tissues (the penalty)
            future_values_func = interp1d(BEDT, (Values[state - 1] * prob).sum(axis=1))
            future_values = future_values_func(
                BED_calc0(actionspace, abt)
            )  # for each action and sparing factor calculate the penalty of the action and add the future value we will only have as many future values as we have actions (not sparing dependent)
            Vs = -BEDN + future_values
            policy4 = Vs.argmax(axis=1)
        elif (
            fraction_state == fraction and fraction != number_of_fractions
        ):  # actual fraction
            actionspace_clipped = actionspace[
                0 : max_action(accumulated_dose, actionspace, goal) + 1
            ]
            BEDN = BED_calc_matrix(sparing_factors[-1], abn, actionspace_clipped)
            future_BEDT = accumulated_dose + BED_calc0(actionspace_clipped, abt)
            future_BEDT[future_BEDT > goal] = goal + 1
            penalties = np.zeros(future_BEDT.shape)
            penalties[future_BEDT > goal] = -10000
            future_values_func = interp1d(BEDT, (Values[state - 1] * prob).sum(axis=1))
            future_values = future_values_func(
                future_BEDT
            )  # for each action and sparing factor calculate the penalty of the action and add the future value we will only have as many future values as we have actions (not sparing dependent)
            Vs = -BEDN + future_values + penalties
            policy4 = Vs.argmax(axis=1)
        elif (
            fraction == number_of_fractions
        ):  # in this state no penalty has to be defined as the value is not relevant
            best_action = (
                -1 + np.sqrt(1 + 4 * 1 * (goal - accumulated_dose) / abt)
            ) / (2 * 1**2 / abt)
            if accumulated_dose > goal:
                best_action = 0
            if best_action < min_dose:
                best_action = min_dose
            if best_action > max_dose:
                best_action = max_dose
            last_BEDN = BED_calc0(best_action, abn, sparing_factors[-1])
            policy4 = best_action * 10
        else:
            future_value_prob = (Values[state - 1] * prob).sum(axis=1)
            future_values_func = interp1d(BEDT, future_value_prob)
            for tumor_index, tumor_value in enumerate(
                BEDT
            ):  # this and the next for loop allow us to loop through all states
                actionspace_clipped = actionspace[
                    0 : max_action(tumor_value, actionspace, goal) + 1
                ]  # we only allow the actions that do not overshoot
                BEDN = BED_calc_matrix(
                    sf, abn, actionspace_clipped
                )  # this one could be done outside of the loop and only the clipping would happen inside the loop.
                BED = BED_calc_matrix(np.ones(len(sf)), abt, actionspace_clipped)
                if state != 0:
                    future_BEDT = tumor_value + BED
                    future_BEDT[future_BEDT > goal] = goal + 1
                    future_values = future_values_func(
                        future_BEDT
                    )  # for each action and sparing factor calculate the penalty of the action and add the future value we will only have as many future values as we have actions (not sparing dependent)
                    penalties = np.zeros(future_BEDT.shape)
                    penalties[future_BEDT > goal] = -10000
                    Vs = -BEDN + future_values + penalties
                    if Vs.size == 0:
                        best_action = np.zeros(len(sf))
                        valer = np.zeros(len(sf))
                    else:
                        best_action = Vs.argmax(axis=1)
                        valer = Vs.max(axis=1)
                else:  # last state no more further values to add
                    best_action = (
                        -1 + np.sqrt(1 + 4 * 1 * (goal - tumor_value) / abt)
                    ) / (2 * 1**2 / abt)
                    if best_action > max_dose:
                        best_action = max_dose
                    if best_action < min_dose:
                        best_action = min_dose
                    last_BEDN = BED_calc0(best_action, abn, sf)
                    future_BEDT = tumor_value + BED_calc0(best_action, abt)
                    underdose_penalty = 0
                    overdose_penalty = 0
                    if future_BEDT < goal:
                        underdose_penalty = (future_BEDT - goal) * 10
                    if future_BEDT > goal:
                        overdose_penalty = -10000
                    valer = (
                        -last_BEDN
                        + underdose_penalty * np.ones(sf.shape)
                        + overdose_penalty * np.ones(sf.shape)
                    )  # gives the value of each action for all sparing factors. elements 0-len(sparingfactors) are the Values for

                policy[state][tumor_index] = best_action
                Values[state][tumor_index] = valer
    if fraction != number_of_fractions:
        physical_dose = actionspace[policy4]
    if fraction == number_of_fractions:
        physical_dose = policy4 / 10
    tumor_dose = BED_calc0(physical_dose, abt)
    OAR_dose = BED_calc0(physical_dose, abn, sparing_factors[-1])
    return [policy, sf, physical_dose, tumor_dose, OAR_dose]


def whole_plan(
    number_of_fractions,
    sparing_factors,
    alpha,
    beta,
    goal,
    abt=10,
    abn=3,
    min_dose=0,
    max_dose=22.3,
    fixed_prob=0,
    fixed_mean=0,
    fixed_std=0,
):
    """
    calculates all doses for a number_of_fractions fraction treatment (with 6 known sparing factors)
    sparing_factors: list or array of number_of_fractions + 1 sparing factors that have been observed.
    To used a fixed probability distribution, change fixed_prob to 1 and add a fixed mean and std.
    If a fixed probability distribution is chosen, alpha and beta are to be chosen arbitrarily as they will not be used
    Parameters
    ----------
    number_of_fractions : integer
        number of fractions that will be delivered.
    sparing_factors : list/array
        list or array of 6 sparing factors that have been observed.
    alpha : float
        alpha hyperparameter of std prior derived from previous patients.
    beta : float
        beta hyperparameter of std prior derived from previous patients.
    goal : float
        prescribed tumor BED
    abt : float
        alpha-beta ratio of tumor. default is 10
    abn : float
        alpha-beta ratio of OAR. default is 3
    min_dose : float
        minimal physical doses to be delivered in one fraction. The doses are aimed at PTV 95. Defaut is 0
    max_dose : float
        maximal physical doses to be delivered in one fraction. The doses are aimed at PTV 95. Default is 22.3
    fixed_prob : int
        this variable is to turn on a fixed probability distribution. If the variable is not used (0), then the probability will be updated. If the variable is turned to 1, the inserted mean and std will be used for a fixed sparing factor distribution
    fixed_mean: float
        mean of the fixed sparing factor normal distribution
    std_fixed: float
        standard deviation of the fixed sparing factor normal distribution
    Returns
    -------
    List with delivered tumor doses, delivered OAR doses and delivered physical doses

    """
    accumulated_tumor_dose = 0
    accumulated_OAR_dose = 0
    physical_doses = np.zeros(number_of_fractions)
    tumor_doses = np.zeros(number_of_fractions)
    OAR_doses = np.zeros(number_of_fractions)
    for looper in range(0, number_of_fractions):
        [policy, sf, physical_dose, tumor_dose, OAR_dose] = value_eval(
            looper + 1,
            number_of_fractions,
            accumulated_tumor_dose,
            sparing_factors[0 : looper + 2],
            alpha,
            beta,
            goal,
            abt,
            abn,
            min_dose,
            max_dose,
            fixed_prob,
            fixed_mean,
            fixed_std,
        )
        accumulated_tumor_dose += tumor_dose
        accumulated_OAR_dose += OAR_dose
        tumor_doses[looper] = tumor_dose
        OAR_doses[looper] = OAR_dose
        physical_doses[looper] = physical_dose
    return [tumor_doses, OAR_doses, physical_doses]


def whole_plan_print(
    number_of_fractions,
    sparing_factors,
    alpha,
    beta,
    goal,
    abt=10,
    abn=3,
    min_dose=0,
    max_dose=22.3,
    fixed_prob=0,
    fixed_mean=0,
    fixed_std=0,
):
    """
    calculates all doses for a number_of_fractions fraction treatment (with 6 known sparing factors)
    sparing_factors: list or array of number_of_fractions + 1 sparing factors that have been observed.
    To used a fixed probability distribution, change fixed_prob to 1 and add a fixed mean and std.
    If a fixed probability distribution is chosen, alpha and beta are to be chosen arbitrarily as they will not be used
    Parameters. This function prints the result and is optimal to be used when working with the script instead of the GUI
    ----------
    number_of_fractions : integer
        number of fractions that will be delivered.
    sparing_factors : list/array
        list or array of 6 sparing factors that have been observed.
    alpha : float
        alpha hyperparameter of std prior derived from previous patients.
    beta : float
        beta hyperparameter of std prior derived from previous patients.
    goal : float
        prescribed tumor BED
    abt : float
        alpha-beta ratio of tumor. default is 10
    abn : float
        alpha-beta ratio of OAR. default is 3
    min_dose : float
        minimal physical doses to be delivered in one fraction. The doses are aimed at PTV 95. Defaut is 0
    max_dose : float
        maximal physical doses to be delivered in one fraction. The doses are aimed at PTV 95. Default is 22.3
    fixed_prob : int
        this variable is to turn on a fixed probability distribution. If the variable is not used (0), then the probability will be updated. If the variable is turned to 1, the inserted mean and std will be used for a fixed sparing factor distribution
    fixed_mean: float
        mean of the fixed sparing factor normal distribution
    std_fixed: float
        standard deviation of the fixed sparing factor normal distribution
    Returns
    -------
    None

    """
    [tumor_doses, OAR_doses, physical_doses] = whole_plan(
        number_of_fractions,
        sparing_factors,
        alpha,
        beta,
        goal,
        abt,
        abn,
        min_dose,
        max_dose,
        fixed_prob,
        fixed_mean,
        fixed_std,
    )
    for i in range(len(physical_doses)):
        print("physical dose delivered in fraction ", i + 1, "  = ", physical_doses[i])
        print("tumor BED", tumor_doses[i])
        print("OAR BED", OAR_doses[i])
    print("accumulated tumor dose: ", np.sum(tumor_doses))
    print("accumulated OAR dose: ", np.sum(OAR_doses))


def single_fraction(
    number_of_fractions,
    sparing_factors,
    accumulated_tumor_dose,
    alpha,
    beta,
    goal,
    abt=10,
    abn=3,
    min_dose=0,
    max_dose=22.3,
    fixed_prob=0,
    fixed_mean=0,
    fixed_std=0,
):  # Vervollständigen
    """
    calculates the actual dose for a number_of_fractions fraction treatment
    sparing_factors: list or array with all sparing factors that have been observed.
    To used a fixed probability distribution, change fixed_prob to 1 and add a fixed mean and std.
    If a fixed probability distribution is chosen, alpha and beta are to be chosen arbitrarily as they will not be used
    Parameters
    ----------
    number_of_fractions : integer
        number of fractions that will be delivered.
    sparing_factors : list/array
        list or array of all sparing factors that have been observed. e.g. list of 3 sparing factors in fraction 2 (planning,fx1,fx2).
    accumulated_tumor_dose : float
        accumulated BED in tumor (from previous fractions) zero in fraction 1.
    alpha : float
        alpha hyperparameter of std prior derived from previous patients.
    beta : float
        beta hyperparameter of std prior derived from previous patients.
    goal: float
        prescribed tumor dose
    abt : float
        alpha-beta ratio of tumor.
    abn : float
        alpha-beta ratio of OAR.
    min_dose : float
        minimal physical doses to be delivered in one fraction. The doses are aimed at PTV 95.
    max_dose : float
        maximal physical doses to be delivered in one fraction. The doses are aimed at PTV 95 .
    fixed_prob : int
        this variable is to turn on a fixed probability distribution. If the variable is not used (0), then the probability will be updated. If the variable is turned to 1, the inserted mean and std will be used for a fixed sparing factor distribution
    fixed_mean: float
        mean of the fixed sparing factor normal distribution
    std_fixed: float
        standard deviation of the fixed sparing factor normal distribution
    Returns
    -------
    None.

    """
    [policy, sf, physical_dose, tumor_dose, OAR_dose] = value_eval(
        len(sparing_factors) - 1,
        number_of_fractions,
        accumulated_tumor_dose,
        sparing_factors,
        alpha,
        beta,
        goal,
        abt,
        abn,
        min_dose,
        max_dose,
        fixed_prob,
        fixed_mean,
        fixed_std,
    )
    print(
        "physical dose delivered in fraction ",
        len(sparing_factors) - 1,
        ": ",
        physical_dose,
    )
    print("tumor BED: ", tumor_dose)
    print("OAR BED", OAR_dose)



# -*- coding: utf-8 -*-
# log settings
LOG_BOOL = 0
LOG_BOOL_LIST = [0,1]
LOG_LEVEL = 1
LOG_LEVEL_LIST = [0,1,2]

# dose, sf
DOSE_STEPSIZE = 0.1
STATE_STEPSIZE = 1
SF_LOW = 0
SF_HIGH = 1.7
SF_STEPSIZE = 0.01
SF_PROB_THRESHOLD = 1e-5
INF_PENALTY = 1e4

# keys
ALPHA_BETA_TUMOR = 10
ALPHA_BETA_OAR = 3

FULL_DICT = {'number_of_fractions':None,
        'fraction': 0,
        'sparing_factors': None,
        'prob_update': 0,
        'fixed_mean': None,
        'fixed_std': None,
        'shape': None,
        'scale': None,
        'shape_inv': None,
        'scale_inv': None,
        'tumor_goal': None,
        'oar_limit': None,
        'c': None,
        'abt':ALPHA_BETA_TUMOR,
        'abn':ALPHA_BETA_OAR,
        'accumulated_oar_dose': 0,
        'accumulated_tumor_dose': 0,
        'min_dose': 0,
        'max_dose': -1,
        }

SETTING_DICT = {
        'dose_stepsize': DOSE_STEPSIZE,
        'state_stepsize': STATE_STEPSIZE,
        'sf_low': SF_LOW,
        'sf_high': SF_HIGH,
        'sf_stepsize': SF_STEPSIZE,
        'sf_prob_threshold': SF_PROB_THRESHOLD,
        'inf_penalty': INF_PENALTY,
        'plot_policy': 0,
        'plot_values': 0,
        'plot_remains': 0,
        }

STANDARD_LIST = [
        'number_of_fractions',
        'fraction',
        'sparing_factors',
        'prob_update',
        'fixed_mean',
        'fixed_std',
        'shape',
        'scale',
        'shape_inv',
        'scale_inv',
        'abt',
        'abn',
        'accumulated_oar_dose',
        'accumulated_tumor_dose',
        'min_dose',
        'max_dose',
        ]

OAR_LIST = STANDARD_LIST + ['tumor_goal']

TUMOR_LIST = STANDARD_LIST + ['oar_limit']

FRAC_LIST = STANDARD_LIST + ['tumor_goal', 'c']

TUMOR_OAR_LIST = STANDARD_LIST + ['tumor_goal', 'oar_limit']

KEY_DICT = {
        'oar': OAR_LIST, 'oar_old': OAR_LIST, 
        'tumor': TUMOR_LIST, 'tumor_old': TUMOR_LIST,
        'frac': FRAC_LIST, 'frac_old': FRAC_LIST,
        'tumor_oar': TUMOR_OAR_LIST, 'tumor_oar_old': TUMOR_OAR_LIST
        }


# -*- coding: utf-8 -*-
import numpy as np

def bed_calc0(dose, ab, sf=1):
    """
    calculates the BED for a specific dose

    Parameters
    ----------
    dose : float
        physical dose to be delivered.
    ab : float
        alpha-beta ratio of tissue.
    sf : float, optional
        sparing factor. The default is 1 (tumor).

    Returns
    -------
    BED : float
        BED to be delivered based on dose, sparing factor
        and alpha-beta ratio.

    """
    BED = sf * dose * (1 + (sf * dose) / ab)
    return BED


def bed_calc_matrix(actionspace, ab, sf):
    """
    calculates the BED for an array of values

    Parameters
    ----------
    actionspace : list/array
        doses to be delivered.
    ab : float
        alpha-beta ratio of tissue.
    sf : list/array
        list of sparing factors to calculate the correspondent BED.

    Returns
    -------
    BED : List/array
        list of all future BEDs based on the delivered doses
        and sparing factors.

    """
    BED = np.outer(actionspace, sf) * (
        1 + np.outer(actionspace, sf) / ab
    )  # produces a actions space x sparing factor array
    return BED

def convert_to_physical(bed, ab, sf=1):
    """
    Converts given BED to the physical dose

    Parameters
    ----------
    BED : float/array
        tumor or OAR BED for which physical dose has to be calculated.
    ab : float
        alpha-beta ratio.
    sf : float/array
        sparing factor, only specify when OAR BED

    Returns
    -------
    dose : positive values float/array
        physical dose
    """
    bed_array = np.array(bed)
    bed_array[bed_array < 0] = 0
    physical_dose = (-sf + np.sqrt(sf**2 + 4 * sf**2 * bed_array / ab)) / (
        2 * sf**2 / ab)
    return physical_dose



# -*- coding: utf-8 -*-
import numpy as np
import matplotlib.pyplot as plt
from matplotlib.colors import Normalize as normalise
import matplotlib.cm as cm

def plot_val(sfs, states, policies, fractions):
    """
    creates a subplot grid for the policy in each fraction that is given

    Parameters
    ----------
    sfs : array
        1d array with dimension n
    states : array
        1d array with dimension m
    policies : array
        2d array with dimension n_policies:n:m

    Returns
    -------
    fig : matplotlib.pyplot.figure
        matplotlib pyplot figure

    """
    [n_policies, _, _] = policies.shape
    # search for optimal rectangular size of subplot grid
    n_rows = n_columns = int(np.sqrt(n_policies))
    while n_rows * n_columns < n_policies:
        if n_rows >= n_columns:
            n_columns += 1
        else:
            n_rows += 1
    # initiate plot and parameters
    fig, ax = plt.subplots(n_rows, n_columns)
    x_min, x_max, y_min, y_max = sfs[0], sfs[-1], states[0], states[-1]

    # create shared colorbar
    colmin, colmax = np.min(policies), np.max(policies)
    normaliser = normalise(colmin, colmax)
    colormap = cm.get_cmap('turbo')
    im = cm.ScalarMappable(cmap=colormap, norm=normaliser)

    # loop through the axes
    try:
        axs = ax.ravel()
    except:
        # in case ax is a 1x1 subplot
        axs = np.array([ax])
        
    for i, pol in enumerate(policies):
        axs[i].imshow(pol, interpolation=None, origin='upper',
            norm=normaliser, cmap=colormap, aspect='auto',
            extent=[x_min, x_max, y_min, y_max])
        axs[i].set_title(fractions[i])
            
    fig.supxlabel('sparing factor')
    fig.supylabel('remaining BED')
    fig.tight_layout()
    fig.colorbar(mappable=im, ax=axs.tolist())

    return fig

def show_plot():
    plt.show()


# -*- coding: utf-8 -*-
"""
whole plan 3D interpolation. This algorithm tracks tumor and OAR BED. If the prescribed tumor dose can be reached, the OAR dose is minimized. If the prescribed tumor dose can not be reached while staying below
maximum BED, the tumor dose is maximized. The value_eval function calculates the optimal dose for one sparing factor given a sparing factor list and the alpha and beta hyperparameter of previous data (can be calculated with data_fit).
the whole_plan function calculates the whole plan given all sparing factors and the hyperparameters.
For extended functions inspect value_eval or whole_plan. Also read the extended function in the readme file.
The value eval function is build by assigning a penalty depending on how much dose has been delivered to the OAR (in each fraction) and by how far we are from the prescribed dose after the last fractions.

"""

import numpy as np
from scipy.interpolate import RegularGridInterpolator
from scipy.stats import gamma, truncnorm

# right now once 90 is hit it doesnt seem to matter how much is overdosed. somehow this must be fixed


def data_fit(data):
    """
    This function fits the alpha and beta value for the prior

    Parameters
    ----------
    data : array
        a nxk matrix with n the amount of patints and k the amount of sparing factors per patient.

    Returns
    -------
    list
        alpha and beta hyperparameter.
    """
    std = data.std(axis=1)
    alpha, loc, beta = gamma.fit(std, floc=0)
    return [alpha, beta]


def get_truncated_normal(mean=0, sd=1, low=0.01, upp=10):
    """
    produces a truncated normal distribution

    Parameters
    ----------
    mean : float, optional
        The default is 0.
    sd : float, optional
        The default is 1.
    low : float, optional
        The default is 0.01.
    upp : float, optional
        The default is 10.

    Returns
    -------
    scipy.stats._distn_infrastructure.rv_frozen
        distribution function.

    """
    return truncnorm((low - mean) / sd, (upp - mean) / sd, loc=mean, scale=sd)


def probdist(X):
    """
    This function produces a probability distribution based on the normal distribution X

    Parameters
    ----------
    X : scipy.stats._distn_infrastructure.rv_frozen
        distribution function.

    Returns
    -------
    prob : list
        list with probabilities for each sparing factor.

    """
    prob = np.zeros(170)
    idx = 0
    for i in np.arange(0.01, 1.71, 0.01):
        prob[idx] = X.cdf(i + 0.004999999999999999999) - X.cdf(i - 0.005)
        idx += 1
    return prob


def std_calc(measured_data, alpha, beta):
    """
    calculates the most likely standard deviation for a list of k sparing factors and a gamma prior
    measured_data: list/array with k sparing factors

    Parameters
    ----------
    measured_data : list/array
        list/array with k sparing factors
    alpha : float
        shape of gamma distribution
    beta : float
        scale of gamma distrinbution

    Returns
    -------
    std : float
        most likely std based on the measured data and gamma prior

    """
    n = len(measured_data)
    std_values = np.arange(0.00001, 0.5, 0.00001)
    likelihood_values = np.zeros(len(std_values))
    for index, value in enumerate(std_values):
        likelihood_values[index] = (
            value ** (alpha - 1)
            / value ** (n - 1)
            * np.exp(-1 / beta * value)
            * np.exp(-np.var(measured_data) / (2 * (value**2 / n)))
        )
    std = std_values[np.argmax(likelihood_values)]
    return std


def argfind(searched_list, value):
    """
    This function is used to find the index of certain values.
    searched_list: list/array with values
    value: value that should be inside the list
    return: index of value

    Parameters
    ----------
    searched_list : list/array
        list in which our searched value is.
    value : float
        item inside list.

    Returns
    -------
    index : integer
        index of value inside list.

    """
    index = min(range(len(searched_list)), key=lambda i: abs(searched_list[i] - value))
    return index


def BED_calc0(dose, ab, sparing=1):
    """
    calculates the BED for a specific dose

    Parameters
    ----------
    dose : float
        physical dose to be delivered.
    ab : float
        alpha-beta ratio of tissue.
    sparing : float, optional
        sparing factor. The default is 1 (tumor).

    Returns
    -------
    BED : float
        BED to be delivered based on dose, sparing factor and alpha-beta ratio.

    """
    BED = sparing * dose * (1 + (sparing * dose) / ab)
    return BED


def BED_calc_matrix(actionspace, ab, sf):
    """
    calculates the BED for an array of values

    Parameters
    ----------
    sf : list/array
        list of sparing factors to calculate the correspondent BED.
    ab : float
        alpha-beta ratio of tissue.
    actionspace : list/array
        doses to be delivered.

    Returns
    -------
    BED : List/array
        list of all future BEDs based on the delivered doses and sparing factors.

    """
    BED = np.outer(sf, actionspace) * (
        1 + np.outer(sf, actionspace) / ab
    )  # produces a sparing factors x actions space array
    return BED


def value_eval(
    fraction,
    number_of_fractions,
    BED_OAR,
    BED_tumor,
    sparing_factors,
    abt,
    abn,
    bound_OAR,
    bound_tumor,
    alpha,
    beta,
    min_dose=0,
    max_dose=22.3,
    fixed_prob=0,
    fixed_mean=0,
    fixed_std=0,
):
    """
    Calculates the optimal dose for the desired fraction.
    fraction: number of actual fraction (1 for first, 2 for second, etc.)
    To used a fixed probability distribution, change fixed_prob to 1 and add a fixed mean and std.
    If a fixed probability distribution is chosen, alpha and beta are to be chosen arbitrarily as they will not be used
    Parameters
    ----------
    fraction : integer
        number of actual fraction (1 for first, 2 for second, etc.).
    number_of_fractions : integer
        number of fractions that will be delivered.
    BED_OAR : float
        accumulated BED in OAR (from previous fractions) zero in fraction 1.
    BED_tumor : float
        accumulated BED in tumor (from previous fractions) zero in fraction 1.
    sparing_factors : TYPE
        list or array of all sparing factors that have been observed. e.g. list of 3 sparing factors in fraction 2 (planning,fx1,fx2).
    abt : float
        alpha-beta ratio of tumor.
    abn : float
        alpha-beta ratio of OAR.
    bound_OAR : float
        maximal BED of OAR
    bound_tumor : float
        prescribed tumor BED.
    alpha : float
        alpha hyperparameter of std prior derived from previous patients.
    beta : float
        beta hyperparameter of std prior derived from previous patients
    min_dose : float
        minimal physical doses to be delivered in one fraction. The doses are aimed at PTV 95
    max_dose : float
        maximal physical doses to be delivered in one fraction. The doses are aimed at PTV 95
    fixed_prob : int
        this variable is to turn on a fixed probability distribution. If the variable is not used (0), then the probability will be updated. If the variable is turned to 1, the inserted mean and std will be used for a fixed sparing factor distribution
    fixed_mean: float
        mean of the fixed sparing factor normal distribution
    std_fixed: float
        standard deviation of the fixed sparing factor normal distribution
    Returns
    -------
    list
        list with following arrays/values:
        physical_dose: optimal physical dose for actual fraction
        accumulated_tumor_dose: accumulated tumor BED
        accumulated_OAR_dose: accumulated OAR BED
        tumor_dose: tumor BED to be delivered
        OAR_dose: OAR BED to be delivered

    """

    if fixed_prob != 1:
        mean = np.mean(
            sparing_factors
        )  # extract the mean and std to setup the sparingfactor distribution
        standard_deviation = std_calc(sparing_factors, alpha, beta)
    if fixed_prob == 1:
        mean = fixed_mean
        standard_deviation = fixed_std
    X = get_truncated_normal(mean=mean, sd=standard_deviation, low=0, upp=1.3)
    prob = np.array(probdist(X))
    sf = np.arange(0.01, 1.71, 0.01)
    sf = sf[prob > 0.00001]  # get rid of all probabilities below 10^-5
    prob = prob[prob > 0.00001]
    underdosepenalty = 10
    BEDT = np.arange(BED_tumor, bound_tumor, 1)  # tumordose
    BEDNT = np.arange(BED_OAR, bound_OAR, 1)  # OAR dose
    BEDNT = np.concatenate((BEDNT, [bound_OAR, bound_OAR + 1]))
    BEDT = np.concatenate((BEDT, [bound_tumor, bound_tumor + 1]))
    Values = np.zeros(
        [(number_of_fractions - fraction), len(BEDT), len(BEDNT), len(sf)]
    )  # 2d values list with first indice being the BED and second being the sf
    if max_dose > (-1 + np.sqrt(1 + 4 * 1 * (bound_tumor) / abt)) / (
        2 * 1**2 / abt
    ):  # we constrain the maximum dose so that no more dose than what is needed would be checked in the actionspace
        max_dose = (-1 + np.sqrt(1 + 4 * 1 * (bound_tumor) / abt)) / (2 * 1**2 / abt)
    if min_dose > max_dose:
        min_dose = max_dose - 0.1
    actionspace = np.arange(min_dose, max_dose + 0.1, 0.1)
    policy = np.zeros(
        ((number_of_fractions - fraction), len(BEDT), len(BEDNT), len(sf))
    )
    upperbound_normal_tissue = bound_OAR + 1
    upperbound_tumor = bound_tumor + 1

    OAR_dose = BED_calc_matrix(
        actionspace, abn, sf
    )  # calculates the dose that is deposited into the normal tissue for all sparing factors
    tumor_dose = BED_calc_matrix(actionspace, abt, 1)[
        0
    ]  # this is the dose delivered to the tumor
    actual_fraction_sf = argfind(sf, np.round(sparing_factors[-1], 2))

    for index, frac_state_plus in enumerate(
        np.arange(number_of_fractions + 1, fraction, -1)
    ):  # We have five fractionations with 2 special cases 0 and 4
        frac_state = frac_state_plus - 1
        if (
            frac_state == 1
        ):  # first state with no prior dose delivered so we dont loop through BEDNT
            future_OAR = BED_OAR + OAR_dose[actual_fraction_sf]
            future_tumor = BED_tumor + tumor_dose
            future_OAR[
                future_OAR > bound_OAR
            ] = upperbound_normal_tissue  # any dose surpassing the upper bound will be set to the upper bound which will be penalized strongly
            future_tumor[future_tumor > bound_tumor] = upperbound_tumor
            future_values_prob = (Values[index - 1] * prob).sum(
                axis=2
            )  # future values of tumor and oar state
            value_interpolation = RegularGridInterpolator(
                (BEDT, BEDNT), future_values_prob
            )
            future_value_actual = value_interpolation(
                np.array([future_tumor, future_OAR]).T
            )
            Vs = future_value_actual - OAR_dose[actual_fraction_sf]
            actual_policy = Vs.argmax(axis=0)

        elif (
            frac_state == fraction
        ):  # if we are in the actual fraction we do not need to check all possible BED states but only the one we are in
            if fraction != number_of_fractions:
                future_OAR = BED_OAR + OAR_dose[actual_fraction_sf]
                overdosing = (future_OAR - bound_OAR).clip(min=0)
                future_OAR[
                    future_OAR > bound_OAR
                ] = upperbound_normal_tissue  # any dose surpassing the upper bound will be set to the upper bound which will be penalized strongly
                future_tumor = BED_tumor + tumor_dose
                future_tumor[future_tumor > bound_tumor] = upperbound_tumor
                future_values_prob = (Values[index - 1] * prob).sum(
                    axis=2
                )  # future values of tumor and oar state
                value_interpolation = RegularGridInterpolator(
                    (BEDT, BEDNT), future_values_prob
                )
                penalties = (
                    overdosing * -10000000000
                )  # additional penalty when overdosing is needed when choosing a minimum dose to be delivered
                future_value_actual = value_interpolation(
                    np.array([future_tumor, future_OAR]).T
                )
                Vs = future_value_actual - OAR_dose[actual_fraction_sf] + penalties
                actual_policy = Vs.argmax(axis=0)
            else:
                sf_end = sparing_factors[-1]
                best_action_BED = (
                    -sf_end
                    + np.sqrt(
                        sf_end**2 + 4 * sf_end**2 * (bound_OAR - BED_OAR) / abn
                    )
                ) / (2 * sf_end**2 / abn)
                best_action_tumor = (
                    -1 + np.sqrt(1 + 4 * 1 * (bound_tumor - BED_tumor) / abt)
                ) / (2 * 1**2 / abt)
                best_action = np.min([best_action_BED, best_action_tumor], axis=0)
                if BED_OAR > bound_OAR or BED_tumor > bound_tumor:
                    best_action = np.ones(best_action.shape) * min_dose
                if best_action > max_dose:
                    best_action = max_dose
                if best_action < min_dose:
                    best_action = min_dose

                future_tumor = BED_tumor + BED_calc0(best_action, abt)
                future_OAR = BED_OAR + BED_calc0(best_action, abn, sf_end)
                actual_policy = best_action * 10
        else:
            future_values_prob = (Values[index - 1] * prob).sum(
                axis=2
            )  # future values of tumor and oar state
            value_interpolation = RegularGridInterpolator(
                (BEDT, BEDNT), future_values_prob
            )  # interpolation function
            for tumor_index, tumor_value in enumerate(BEDT):
                for OAR_index, OAR_value in enumerate(
                    BEDNT
                ):  # this and the next for loop allow us to loop through all states
                    future_OAR = OAR_dose + OAR_value
                    overdosing = (future_OAR - bound_OAR).clip(min=0)
                    future_OAR[
                        future_OAR > bound_OAR
                    ] = upperbound_normal_tissue  # any dose surpassing 90.1 is set to 90.1
                    future_tumor = tumor_value + tumor_dose
                    future_tumor[
                        future_tumor > bound_tumor
                    ] = upperbound_tumor  # any dose surpassing the tumor bound is set to tumor_bound + 0.1

                    if (
                        frac_state == number_of_fractions
                    ):  # last state no more further values to add
                        best_action_BED = (
                            -sf
                            + np.sqrt(
                                sf**2 + 4 * sf**2 * (bound_OAR - OAR_value) / abn
                            )
                        ) / (
                            2 * sf**2 / abn
                        )  # calculate maximal dose that can be delivered to OAR and tumor
                        best_action_tumor = (
                            -np.ones(len(sf))
                            + np.sqrt(
                                np.ones(len(sf))
                                + 4
                                * np.ones(len(sf))
                                * (bound_tumor - tumor_value)
                                / abt
                            )
                        ) / (2 * np.ones(len(sf)) ** 2 / abt)
                        best_action = np.min(
                            [best_action_BED, best_action_tumor], axis=0
                        )  # take the smaller of both doses to not surpass the limit
                        best_action[best_action > max_dose] = max_dose
                        best_action[best_action < min_dose] = min_dose
                        if (
                            OAR_value > bound_OAR or tumor_value > bound_tumor
                        ):  # if the limit is already surpassed we add a penaltsy
                            best_action = np.ones(best_action.shape) * min_dose
                        future_OAR = OAR_value + BED_calc0(best_action, abn, sf)
                        future_tumor = tumor_value + BED_calc0(best_action, abt, 1)
                        overdose_penalty2 = np.zeros(
                            best_action.shape
                        )  # we need a second penalty if we overdose in the last fraction
                        overdose_penalty3 = np.zeros(best_action.shape)
                        overdose_penalty2[
                            future_tumor > bound_tumor + 0.0001
                        ] = -100000000000
                        overdose_penalty3[
                            future_OAR > bound_OAR + 0.0001
                        ] = (
                            -100000000000
                        )  # A small number has to be added as sometimes 90. > 90 was True
                        end_penalty = (
                            -abs(future_tumor - bound_tumor) * underdosepenalty
                        )  # the farther we are away from the prescribed dose, the higher the penalty. Under- and overdosing is punished
                        end_penalty_OAR = (
                            -(future_OAR - bound_OAR).clip(min=0) * 1000
                        )  # if overdosing the OAR is not preventable, the overdosing should stay as low as possible
                        Values[index][tumor_index][OAR_index] = (
                            end_penalty
                            - BED_calc0(best_action, abn, sf)
                            + overdose_penalty2
                            + overdose_penalty3
                            + end_penalty_OAR
                        )  # we also substract all the dose delivered to the OAR so the algorithm tries to minimize it
                        policy[index][tumor_index][OAR_index] = best_action * 10
                    else:
                        future_value = np.zeros([len(sf), len(actionspace)])
                        for actual_sf in range(0, len(sf)):
                            future_value[actual_sf] = value_interpolation(
                                np.array([future_tumor, future_OAR[actual_sf]]).T
                            )
                        penalties = (
                            overdosing * -10000000000
                        )  # additional penalty when overdosing is needed when choosing a minimum dose to be delivered
                        Vs = future_value - OAR_dose + penalties
                        best_action = Vs.argmax(axis=1)
                        valer = Vs.max(axis=1)
                        policy[index][tumor_index][OAR_index] = best_action
                        Values[index][tumor_index][OAR_index] = valer
    if fraction != number_of_fractions:
        physical_dose = actionspace[actual_policy]
    else:
        physical_dose = actual_policy / 10
    tumor_dose = BED_calc0(physical_dose, abt)
    OAR_dose = BED_calc0(physical_dose, abn, sparing_factors[-1])
    accumulated_tumor_dose = BED_calc0(physical_dose, abt) + BED_tumor
    accumulated_OAR_dose = BED_calc0(physical_dose, abn, sparing_factors[-1]) + BED_OAR
    return [
        physical_dose,
        accumulated_tumor_dose,
        accumulated_OAR_dose,
        tumor_dose,
        OAR_dose,
    ]


def whole_plan(
    number_of_fractions,
    sparing_factors,
    abt,
    abn,
    bound_OAR,
    bound_tumor,
    alpha,
    beta,
    min_dose=0,
    max_dose=22.3,
    fixed_prob=0,
    fixed_mean=0,
    std_fixed=0,
):
    """
    calculates all doses for a number_of_fractions fraction treatment (with 6 known sparing factors)
    sparing_factors: list or array of number_of_fractions + 1 sparing factors that have been observed.
    To used a fixed probability distribution, change fixed_prob to 1 and add a fixed mean and std.
    If a fixed probability distribution is chosen, alpha and beta are to be chosen arbitrarily as they will not be used
    Parameters
    ----------
    sparing_factors : list/array
        list or array of 6 sparing factors that have been observed.
    number_of_fractions : integer
        number of fractions that will be delivered.
    abt : float
        alpha-beta ratio of tumor.
    abn : float
        alpha-beta ratio of OAR.
    bound_OAR : float
        maximal BED of OAR.
    bound_tumor : float
        prescribed tumor BED.
    alpha : float
        alpha hyperparameter of std prior derived from previous patients.
    beta : float
        beta hyperparameter of std prior derived from previous patients.
    min_dose : float
        minimal physical doses to be delivered in one fraction. The doses are aimed at PTV 95
    max_dose : float
        maximal physical doses to be delivered in one fraction. The doses are aimed at PTV 95
    fixed_prob : int
        this variable is to turn on a fixed probability distribution. If the variable is not used (0), then the probability will be updated. If the variable is turned to 1, the inserted mean and std will be used for a fixed sparing factor distribution
    fixed_mean: float
        mean of the fixed sparing factor normal distribution
    std_fixed: float
        standard deviation of the fixed sparing factor normal distribution
    Returns
    -------
    List with delivered tumor doses, delivered OAR doses and delivered physical doses

    """
    physical_doses = np.zeros(number_of_fractions)
    tumor_doses = np.zeros(number_of_fractions)
    OAR_doses = np.zeros(number_of_fractions)
    accumulated_OAR_dose = 0
    accumulated_tumor_dose = 0
    for looper in range(0, number_of_fractions):
        [
            actual_policy,
            accumulated_tumor_dose,
            accumulated_OAR_dose,
            tumor_dose,
            OAR_dose,
        ] = value_eval(
            looper + 1,
            number_of_fractions,
            accumulated_OAR_dose,
            accumulated_tumor_dose,
            sparing_factors[0 : looper + 2],
            abt,
            abn,
            bound_OAR,
            bound_tumor,
            alpha,
            beta,
            min_dose,
            max_dose,
            fixed_prob,
            fixed_mean,
            std_fixed,
        )
        physical_doses[looper] = actual_policy
        tumor_doses[looper] = tumor_dose
        OAR_doses[looper] = OAR_dose
        print("calcultions fraction ", looper + 1, " done")
    return [tumor_doses, OAR_doses, physical_doses]


def whole_plan_print(
    number_of_fractions,
    sparing_factors,
    abt,
    abn,
    bound_OAR,
    bound_tumor,
    alpha,
    beta,
    min_dose=0,
    max_dose=22.2,
    fixed_prob=0,
    fixed_mean=0,
    fixed_std=0,
):
    """
    calculates all doses for a 5 fraction treatment (with 6 known sparing factors)
    sparing_factors: list or array of 6 sparing factors that have been observed.
    To used a fixed probability distribution, change fixed_prob to 1 and add a fixed mean and std.
    If a fixed probability distribution is chosen, alpha and beta are to be chosen arbitrarily as they will not be used. This function prints the result and is optimal to be used when working with the script instead of the GUI
    Parameters
    ----------
    sparing_factors : list/array
        list or array of 6 sparing factors that have been observed.
    number_of_fractions : integer
        number of fractions that will be delivered.
    abt : float
        alpha-beta ratio of tumor.
    abn : float
        alpha-beta ratio of OAR.
    bound_OAR : float
        maximal BED of OAR.
    bound_tumor : float
        prescribed tumor BED.
    alpha : float
        alpha hyperparameter of std prior derived from previous patients.
    beta : float
        beta hyperparameter of std prior derived from previous patients.
    min_dose : float
        minimal physical doses to be delivered in one fraction. The doses are aimed at PTV 95
    max_dose : float
        maximal physical doses to be delivered in one fraction. The doses are aimed at PTV 95
    fixed_prob : int
        this variable is to turn on a fixed probability distribution. If the variable is not used (0), then the probability will be updated. If the variable is turned to 1, the inserted mean and std will be used for a fixed sparing factor distribution
    fixed_mean: float
        mean of the fixed sparing factor normal distribution
    std_fixed: float
        standard deviation of the fixed sparing factor normal distribution
    Returns
    -------
    None.

    """
    [tumor_doses, OAR_doses, physical_doses] = whole_plan(
        number_of_fractions,
        sparing_factors,
        abt,
        abn,
        bound_OAR,
        bound_tumor,
        alpha,
        beta,
        min_dose,
        max_dose,
        fixed_prob,
        fixed_mean,
        fixed_std,
    )
    for i in range(number_of_fractions):
        print("fraction ", i + 1)
        print("physical dose delivered = ", np.round(physical_doses[i], 2))
        print("tumor BED delivered = ", np.round(tumor_doses[i], 2))
        print("OAR BED delivered = ", np.round(OAR_doses[i], 2))
    print("total tumor BED = ", np.sum(tumor_doses))
    print("total OAR BED = ", np.sum(OAR_doses))


def single_fraction_print(
    number_of_fractions,
    sparing_factors,
    BED_OAR,
    BED_tumor,
    abt,
    abn,
    bound_OAR,
    bound_tumor,
    alpha,
    beta,
    min_dose,
    max_dose,
    fixed_prob=0,
    fixed_mean=0,
    fixed_std=0,
):
    """
    calculates the actual dose for a number_of_fractions fraction treatment
    sparing_factors: list or array with all sparing factors that have been observed.
    To used a fixed probability distribution, change fixed_prob to 1 and add a fixed mean and std.
    If a fixed probability distribution is chosen, alpha and beta are to be chosen arbitrarily as they will not be used
    Parameters
    ----------
    sparing_factors : list/array
        list or array of all sparing factors that have been observed. e.g. list of 3 sparing factors in fraction 2 (planning,fx1,fx2).
    number_of_fractions : integer
        number of fractions that will be delivered.
    BED_OAR : float
        accumulated BED in OAR (from previous fractions) zero in fraction 1.
    BED_tumor : float
        accumulated BED in tumor (from previous fractions) zero in fraction 1.
    abt : float
        alpha-beta ratio of tumor.
    abn : float
        alpha-beta ratio of OAR.
    bound_OAR : float
        maximal BED of OAR
    bound_tumor : float
        prescribed tumor BED.
    alpha : float
        alpha hyperparameter of std prior derived from previous patients.
    beta : float
        beta hyperparameter of std prior derived from previous patients.
    min_dose : float
        minimal physical doses to be delivered in one fraction. The doses are aimed at PTV 95.
    max_dose : float
        maximal physical doses to be delivered in one fraction. The doses are aimed at PTV 95 .
    fixed_prob : int
        this variable is to turn on a fixed probability distribution. If the variable is not used (0), then the probability will be updated. If the variable is turned to 1, the inserted mean and std will be used for a fixed sparing factor distribution
    fixed_mean: float
        mean of the fixed sparing factor normal distribution
    std_fixed: float
        standard deviation of the fixed sparing factor normal distribution
    Returns
    -------
    None.

    """
    [
        actual_policy,
        accumulated_tumor_dose,
        accumulated_OAR_dose,
        tumor_dose,
        OAR_dose,
    ] = value_eval(
        len(sparing_factors) - 1,
        number_of_fractions,
        BED_OAR,
        BED_tumor,
        sparing_factors,
        abt,
        abn,
        bound_OAR,
        bound_tumor,
        alpha,
        beta,
        min_dose,
        max_dose,
        fixed_prob,
        fixed_mean,
        fixed_std,
    )
    print("fraction ", len(sparing_factors) - 1)
    print("physical dose delivered = ", actual_policy)
    print("tumor BED delivered = ", tumor_dose)
    print("OAR BED delivered = ", OAR_dose)



# -*- coding: utf-8 -*-
import time
from adaptfx import aft_message, aft_error, aft_warning, aft_message
nme = __name__

def stat_rounding(number, decimal):
    magnitude = 10 ** decimal
    return round(number * magnitude) / magnitude

def timing(start=None):
    """
    measure time for general process:
    for start: var_name = timing()
    for strop: timing(var_name)

    Parameters
    ----------
    start : time of starting function
        the default is None.

    Returns
    -------
    start : float
        starting time
        
    """
    if start == None:
        start_time = time.perf_counter()
        return start_time
    else:
        stop = time.perf_counter()
        time_elapsed = stat_rounding((stop - start), 4)
        aft_message(f'process duration: {time_elapsed} s:', nme, 1)

def key_reader(all_keys, full_dict, user_keys, algorithm):
    """
    read and check all keys from a instruction
    file by cycling through all_keys

    Parameters
    ----------
    all_keys : dict
        all keys necessary for some algorithm type.
    full_dict : dict
        all possible keys.
    user_keys : dict
        read in keys from an instruction file.
    algorithm : string
        type of algorithm.

    Returns
    -------
    whole dict : dict
        all keys copied from parameters
        
    """
    aft_message('loading keys...', nme, 1)
    whole_dict = full_dict.copy()
    key_dict = all_keys[algorithm]

    for key in key_dict:
        if key in user_keys:
            whole_dict[key] = user_keys[key]
        elif key not in user_keys:
            if full_dict[key] == None:
                aft_error(f'missing mandatory key: "{key}"', nme)
            else:
                whole_dict[key] = full_dict[key]

    for key in user_keys:
        if key not in key_dict and key in full_dict:
            aft_warning(
                f'key: "{key}" is not allowed for "{algorithm}"', nme, 0
                )
        elif key not in full_dict:
            aft_warning(f'unexpected key: "{key}"', nme, 0)

    return dict((k, whole_dict[k]) for k in key_dict)

def setting_reader(all_settings, user_settings):
    """
    read and check all keys from a settings
    by cycling through all_settings

    Parameters
    ----------
    all_settings : dict
        all settings necessary for calculation
    settings : dict
        all settings for calculation

    Returns
    -------
    whole_settings : dict
        all settings
        
    """
    aft_message('loading settings...', nme, 1)
    whole_settings = all_settings.copy()

    for skey in all_settings:
        if skey in user_settings:
            whole_settings[skey] = user_settings[skey]
    
    for skey in user_settings:
        if skey not in all_settings:
            aft_warning(f'unexpected setting: "{skey}"', nme, 0)

    return whole_settings

class DotDict(dict):
    """
    object with nested dot.notation to
    access dictionary attributes
    """
    MARKER = object()

    def __init__(self, value=None):
        if value is None:
            pass
        elif isinstance(value, dict):
            for key in value:
                self.__setitem__(key, value[key])
        else:
            raise TypeError('expected dict')

    def __setitem__(self, key, value):
        if isinstance(value, dict) and not isinstance(value, DotDict):
            value = DotDict(value)
        super(DotDict, self).__setitem__(key, value)

    def __getitem__(self, key):
        return self.get(key)
    
    def __delitem__(self, key):
        # found = self.get(key, DotDict.MARKER)
        # if found is DotDict.MARKER:
        super(DotDict, self).__delitem__(key)

    __setattr__, __getattr__, __delattr__ = __setitem__, __getitem__, __delitem__


# -*- coding: utf-8 -*-
import sys
import logging
import os

prompt = 'AFT> '
empty = ''

def logging_init(filename, log, level):
    """
    log initialisation to write to filename

    Parameters
    ----------
    filename : string
        filename of log file
    log : bool
        if true store log to filename
    debug: bool
        if true log extensive message

    Returns
    -------
    None
        
    """
    if level == 0:
        format_file = '%(asctime)s [%(levelname)s] [%(name)s]: %(message)s'
        format_out = prompt + '[%(levelname)s] [%(name)s]: %(message)s'
        log_level = logging.DEBUG
    elif level == 1:
        format_file = '%(asctime)s [%(levelname)s]: %(message)s'
        format_out = prompt + '[%(levelname)s]: %(message)s'
        log_level = logging.INFO
    elif level == 2:
        format_file = '%(asctime)s [%(levelname)s]: %(message)s'
        format_out = '[%(levelname)s]: %(message)s'
        log_level = logging.ERROR
    
    if log:
        logfile_extension = "log"
        # create logfile name
        # get the basename before .json extension
        basename = filename.rsplit('.')[0]
        # search for existing filename ...
        i = 1
        while os.path.exists(f'{basename}_{i}.{logfile_extension}'):
            # exponential search if many files exist
            i *= 2
        a, b = (i // 2, i)
        while a+1 < b:
            c = (a + b) // 2
            if os.path.exists(f'{basename}_{c}.{logfile_extension}'):
                a, b = (c, b)
            else:
                a, b = (a, c)
        # ... end of search
        log_filename = f'{basename}_{b}.{logfile_extension}'
        logging.basicConfig(
            format=format_file,
            level=log_level, 
            datefmt='%d-%m-%Y %H:%M:%S',
            handlers=[
                logging.FileHandler(log_filename),
                logging.StreamHandler(sys.stdout)
                ]
        )
    else:
        logging.basicConfig(
            format=format_out,
            level=log_level, 
            handlers=[
                logging.StreamHandler(sys.stdout)
                ]
        )

def aft_error(error, name):
    """
    error message that exits process

    Parameters
    ----------
    error : string
        message that is logged on logging.ERROR level
    name : string
        logger name

    Returns
    -------
    None
        
    """
    log_name = logging.getLogger(name)
    log_name.error(error)
    log_name.info('exiting Session...')
    sys.exit()

def aft_warning(warning, name, mode=0):
    """
    warning message

    Parameters
    ----------
    warning : string
        message that is logged in logging.WARNING level
    name : string
        logger name

    Returns
    -------
    None
        
    """
    log_name = logging.getLogger(name)
    if mode == 1:
        log_name.info(empty)
    log_name.warning(warning)

def aft_message(message, name, mode=0):
    """
    information message

    Parameters
    ----------
    message : string
        message that is logged on logging.INFO level
    name : string
        logger name

    Returns
    -------
    None
        
    """
    log_name = logging.getLogger(name)
    if mode == 1:
        log_name.info(empty)
    log_name.info(message)

def aft_message_info(message, info, name, mode=0):
    """
    information message belonging to certain string

    Parameters
    ----------
    message : string
        message that is logged on logging.INFO level
    info : string
        variable information
    name : string
        logger name

    Returns
    -------
    None
        
    """
    log_name = logging.getLogger(name)
    if mode == 1:
        log_name.info(empty)
    log_name.info(f'{message} {info}')

def aft_message_dict(message, dict, name, mode=0):
    """
    print statement for dictionary

    Parameters
    ----------
    message : string
        message that is logged on logging.INFO level
    dict : dict
        dictionary of parameters
    name : string
        logger name

    Returns
    -------
    None
        
    """
    log_name = logging.getLogger(name)
    if mode == 1:
        log_name.info(empty)
    log_name.info(message)
    for key, value in dict.items():
        log_name.info(f'|{key: <22}| {value}')
        
def aft_message_list(message, struct, name, mode=0):
    """
    print statement for general structures (e.g. list)

    Parameters
    ----------
    message : string
        message that is logged on logging.INFO level
    struct : list
        list of fractionation plans
    name : string
        logger name

    Returns
    -------
    None
        
    """
    log_name = logging.getLogger(name)
    if mode == 1:
        log_name.info(empty)
    log_name.info(message)
    log_name.info('%s', struct)



"""
GUI for 2D adaptive fractionation with minimum and maximum dose
"""
# The whole GUI needs to be class like to have a working scrollbar. Update coming soon!
import tkinter as tk
import tkinter.ttk as ttk
from tkinter import filedialog as fd
from tkinter.messagebox import showinfo

import interpol2D_OAR as intOAR
import interpol2D_tumor as inttumor
import numpy as np
import pandas as pd
from scipy.stats import gamma


class VerticalScrolledFrame(tk.Frame):
    """A pure Tkinter scrollable frame that actually works!
    * Use the 'interior' attribute to place widgets inside the scrollable frame
    * Construct and pack/place/grid normally
    * This frame only allows vertical scrolling

    """

    def __init__(self, parent, *args, **kw):
        tk.Frame.__init__(self, parent, *args, **kw)

        # create a canvas object and a vertical scrollbar for scrolling it
        vscrollbar = ttk.Scrollbar(self, orient=tk.VERTICAL)
        vscrollbar.pack(fill=tk.Y, side=tk.RIGHT, expand=tk.FALSE)
        canvas = tk.Canvas(
            self, bd=0, highlightthickness=0, yscrollcommand=vscrollbar.set, height=1000
        )
        canvas.pack(side=tk.LEFT, fill=tk.BOTH, expand=tk.TRUE)
        vscrollbar.config(command=canvas.yview)
        # reset the view
        canvas.xview_moveto(0)
        canvas.yview_moveto(0)

        # create a frame inside the canvas which will be scrolled with it
        self.interior = interior = tk.Frame(canvas)
        interior_id = canvas.create_window(0, 0, window=interior, anchor=tk.NW)

        # track changes to the canvas and frame width and sync them,
        # also updating the scrollbar
        def _configure_interior(event):
            # update the scrollbars to match the size of the inner frame
            size = (interior.winfo_reqwidth(), interior.winfo_reqheight())
            canvas.config(scrollregion="0 0 %s %s" % size)
            if interior.winfo_reqwidth() != canvas.winfo_width():
                # update the canvas's width to fit the inner frame
                canvas.config(width=interior.winfo_reqwidth())

        interior.bind("<Configure>", _configure_interior)

        def _configure_canvas(event):
            if interior.winfo_reqwidth() != canvas.winfo_width():
                # update the inner frame's width to fill the canvas
                canvas.itemconfigure(interior_id, width=canvas.winfo_width())

        canvas.bind("<Configure>", _configure_canvas)


class GUI2Dextended:
    def __init__(self, master):
        self.master = master
        master.title("2D adaptive fractionation calculator extended")
        self.frame = VerticalScrolledFrame(master)
        self.frame.pack()
        self.frm_probdis = tk.Frame(
            master=self.frame.interior, relief=tk.SUNKEN, borderwidth=3
        )
        self.frm_probdis.pack()
        self.data = []
        self.info_funcs = [self.info1, self.info2, self.info3, self.info4, self.info5]
        self.info_buttons = ["btn_path", "btn_mean", "btn_std", "btn_shae", "btn_scale"]
        for idx in range(len(self.info_funcs)):
            globals()[self.info_buttons[idx]] = tk.Button(
                master=self.frm_probdis, text="?", command=self.info_funcs[idx]
            )
            globals()[self.info_buttons[idx]].grid(row=idx + 1, column=4)
        for idx in range(len(self.info_funcs)):
            globals()[self.info_buttons[idx]] = tk.Button(
                master=self.frm_probdis, text="?", command=self.info_funcs[idx]
            )
            globals()[self.info_buttons[idx]].grid(row=idx + 1, column=4)

        self.var_radio = tk.IntVar()
        self.var_radio.set(1)
        self.hyper_insert = tk.Radiobutton(
            master=self.frm_probdis,
            text="hyperparameters",
            justify="left",
            variable=self.var_radio,
            value=1,
            command=self.checkbox1,
        )
        self.hyper_insert.grid(row=0, column=0)
        self.file_insert = tk.Radiobutton(
            master=self.frm_probdis,
            text="prior data",
            justify="left",
            variable=self.var_radio,
            value=2,
            command=self.checkbox1,
        )
        self.file_insert.grid(row=0, column=1)
        self.fixed_insert = tk.Radiobutton(
            master=self.frm_probdis,
            text="define normal distribution",
            justify="left",
            variable=self.var_radio,
            value=3,
            command=self.checkbox1,
        )
        self.fixed_insert.grid(row=0, column=2)

        # open button
        self.lbl_open = tk.Label(
            master=self.frm_probdis, text="load patient data for prior"
        )
        self.lbl_open.grid(row=1, column=0)
        self.btn_open = tk.Button(
            self.frm_probdis, text="Open a File", command=self.select_file
        )
        self.ent_file = tk.Entry(master=self.frm_probdis, width=20)
        self.btn_open.grid(row=1, column=1)

        self.lbl_mean = tk.Label(
            master=self.frm_probdis, text="mean of normal distribution:"
        )
        self.lbl_mean.grid(row=2, column=0)
        self.ent_mean = tk.Entry(master=self.frm_probdis, width=30)
        self.ent_mean.grid(row=2, column=1, columnspan=2)

        self.lbl_std = tk.Label(
            master=self.frm_probdis, text="std of normal distribution:"
        )
        self.lbl_std.grid(row=3, column=0)
        self.ent_std = tk.Entry(master=self.frm_probdis, width=30)
        self.ent_std.grid(row=3, column=1, columnspan=2)

        self.lbl_alpha = tk.Label(
            master=self.frm_probdis, text="shape of gamma distribution (alpha):"
        )
        self.lbl_alpha.grid(row=4, column=0)
        self.ent_alpha = tk.Entry(master=self.frm_probdis, width=30)
        self.ent_alpha.grid(row=4, column=1, columnspan=2)

        self.lbl_beta = tk.Label(
            master=self.frm_probdis, text="scale of gamma distribution (beta):"
        )
        self.lbl_beta.grid(row=5, column=0)
        self.ent_beta = tk.Entry(master=self.frm_probdis, width=30)
        self.ent_beta.grid(row=5, column=1, columnspan=2)

        self.btn_open.configure(state="disabled")
        self.ent_alpha.configure(state="normal")
        self.ent_beta.configure(state="normal")
        self.ent_file.configure(state="disabled")
        self.ent_mean.configure(state="disabled")
        self.ent_std.configure(state="disabled")
        self.ent_alpha.insert(0, "2.468531215126044")
        self.ent_beta.insert(0, "0.02584178910588476")

        # produce master with extra option like number of fractions.
        self.frm_extras = tk.Frame(
            master=self.frame.interior, relief=tk.SUNKEN, borderwidth=3
        )
        self.frm_extras.pack()

        self.lbl_fractions = tk.Label(
            master=self.frm_extras, text="Total number of fractions"
        )
        self.lbl_fractions.grid(row=0, column=0)
        self.ent_fractions = tk.Entry(master=self.frm_extras, width=30)
        self.ent_fractions.grid(row=0, column=1, columnspan=2)
        self.ent_fractions.insert(0, "5")
        self.btn_infofrac = tk.Button(
            master=self.frm_extras, text="?", command=self.infofrac
        )
        self.btn_infofrac.grid(row=0, column=3)

        self.lbl_mindose = tk.Label(master=self.frm_extras, text="minimum dose")
        self.lbl_mindose.grid(row=1, column=0)
        self.ent_mindose = tk.Entry(master=self.frm_extras, width=30)
        self.ent_mindose.grid(row=1, column=1, columnspan=2)
        self.ent_mindose.insert(0, "0")
        self.btn_mindose = tk.Button(
            master=self.frm_extras, text="?", command=self.infomin
        )
        self.btn_mindose.grid(row=1, column=3)

        self.lbl_maxdose = tk.Label(master=self.frm_extras, text="maximum dose")
        self.lbl_maxdose.grid(row=2, column=0)
        self.ent_maxdose = tk.Entry(master=self.frm_extras, width=30)
        self.ent_maxdose.grid(row=2, column=1, columnspan=2)
        self.ent_maxdose.insert(0, "22.3")
        self.btn_maxdose = tk.Button(
            master=self.frm_extras, text="?", command=self.infomin
        )
        self.btn_maxdose.grid(row=2, column=3)
        # Create a new frame `frm_form` to contain the Label
        # and Entry widgets for entering variable values
        self.frm_form = tk.Frame(
            master=self.frame.interior, relief=tk.SUNKEN, borderwidth=3
        )
        # Pack the frame into the master
        self.frm_form.pack()

        self.frm_buttons = tk.Frame(master=self.frame.interior)
        self.frm_buttons.pack(fill=tk.X, ipadx=5, ipady=5)

        self.frm_output = tk.Frame(
            master=self.frame.interior, relief=tk.SUNKEN, borderwidth=3
        )
        # add label and entry for filename
        self.label = tk.Label(master=self.frm_form, text="file path of prior patients")
        self.ent_file = tk.Entry(master=self.frm_form, width=50)
        self.label.grid(row=0, column=0, sticky="e")
        self.ent_file.grid(row=0, column=1)
        self.info_funcs = [
            self.info10,
            self.info11,
            self.info12,
            self.info13,
            self.info14,
            self.info15,
        ]
        self.info_buttons = [
            "self.btn_sf",
            "self.btn_abt",
            "self.btn_abn",
            "self.btn_OARlimit",
            "self.btn_tumorlimit",
            "self.btn_BED",
        ]
        # List of field labels
        self.labels = [
            "sparing factors separated by spaces:",
            "alpha-beta ratio of tumor:",
            "alpha-beta ratio of OAR:",
            "OAR limit:",
            "prescribed tumor dose:",
            "accumulated BED in tumor or OAR (depending on the tracked organ):",
        ]
        self.ent_sf = tk.Entry(master=self.frm_form, width=50)
        self.lbl_sf = tk.Label(master=self.frm_form, text=self.labels[0])
        self.example_list = [
            "sparing factors separated by space",
            10,
            3,
            90,
            72,
            "only needed if we calculate the dose for a single fraction",
        ]
        self.lbl_sf.grid(row=0, column=0, sticky="e")
        self.ent_sf.grid(row=0, column=1)
        self.ent_sf.insert(0, f"{self.example_list[0]}")
        self.btn_sf = tk.Button(
            master=self.frm_form, text="?", command=self.info_funcs[0]
        )
        self.btn_sf.grid(row=0, column=2)

        self.ent_abt = tk.Entry(master=self.frm_form, width=50)
        self.lbl_abt = tk.Label(master=self.frm_form, text=self.labels[1])
        self.lbl_abt.grid(row=1, column=0, sticky="e")
        self.ent_abt.grid(row=1, column=1)
        self.ent_abt.insert(0, f"{self.example_list[1]}")
        self.btn_abt = tk.Button(
            master=self.frm_form, text="?", command=self.info_funcs[1]
        )
        self.btn_abt.grid(row=1, column=2)

        self.ent_abn = tk.Entry(master=self.frm_form, width=50)
        self.lbl_abn = tk.Label(master=self.frm_form, text=self.labels[2])
        self.lbl_abn.grid(row=2, column=0, sticky="e")
        self.ent_abn.grid(row=2, column=1)
        self.ent_abn.insert(0, f"{self.example_list[2]}")
        self.btn_abn = tk.Button(
            master=self.frm_form, text="?", command=self.info_funcs[2]
        )
        self.btn_abn.grid(row=2, column=2)

        self.ent_OARlimit = tk.Entry(master=self.frm_form, width=50)
        self.lbl_OARlimit = tk.Label(master=self.frm_form, text=self.labels[3])
        self.lbl_OARlimit.grid(row=3, column=0, sticky="e")
        self.ent_OARlimit.grid(row=3, column=1)
        self.ent_OARlimit.insert(0, f"{self.example_list[3]}")
        self.btn_OARlimit = tk.Button(
            master=self.frm_form, text="?", command=self.info_funcs[3]
        )
        self.btn_OARlimit.grid(row=3, column=2)

        self.ent_tumorlimit = tk.Entry(master=self.frm_form, width=50)
        self.lbl_tumorlimit = tk.Label(master=self.frm_form, text=self.labels[4])
        self.lbl_tumorlimit.grid(row=4, column=0, sticky="e")
        self.ent_tumorlimit.grid(row=4, column=1)
        self.ent_tumorlimit.insert(0, f"{self.example_list[4]}")
        self.btn_tumorlimit = tk.Button(
            master=self.frm_form, text="?", command=self.info_funcs[4]
        )
        self.btn_tumorlimit.grid(row=4, column=2)

        self.ent_BED = tk.Entry(master=self.frm_form, width=50)
        self.lbl_BED = tk.Label(master=self.frm_form, text=self.labels[5])
        self.lbl_BED.grid(row=5, column=0, sticky="e")
        self.ent_BED.grid(row=5, column=1)
        self.ent_BED.insert(0, f"{self.example_list[5]}")
        self.btn_BED = tk.Button(
            master=self.frm_form, text="?", command=self.info_funcs[5]
        )
        self.btn_BED.grid(row=5, column=2)

        self.ent_BED.configure(
            state="disabled"
        )  # the standard calculation is a whole plan calculation
        self.ent_tumorlimit.configure(state="disabled")

        self.btn_compute = tk.Button(
            master=self.frm_buttons, text="compute plan", command=self.compute_plan
        )
        self.btn_compute.pack(side=tk.BOTTOM, ipadx=10)
        self.var = tk.IntVar()
        self.chk_single_fraction = tk.Checkbutton(
            master=self.frm_buttons,
            text="Calculate dose only for actual fraction",
            variable=self.var,
            onvalue=1,
            offvalue=0,
            command=self.checkbox,
        )
        self.chk_single_fraction.pack(side=tk.BOTTOM, padx=10, ipadx=10)
        self.var_OAR = tk.IntVar()
        self.chk_OARswitch = tk.Checkbutton(
            master=self.frm_buttons,
            text="Calculate optimal plan by minimizing OAR and aiming on prescribed tumor dose",
            variable=self.var_OAR,
            onvalue=1,
            offvalue=0,
            command=self.OAR_enable,
        )
        self.chk_OARswitch.pack(side=tk.BOTTOM, padx=10, ipadx=10)

        self.lbl_info = tk.Label(
            master=self.frm_output,
            text="There are several default values set. Only the sparing factors have to been inserted.\nThis program might take some minutes to calculate",
        )
        self.lbl_info.pack()
        self.frm_output.pack(fill=tk.BOTH, ipadx=10, ipady=10)

    def select_file(self):
        filetypes = (("csv files", "*.csv"), ("All files", "*.*"))

        filename = fd.askopenfilename(
            title="Open a file", initialdir="/", filetypes=filetypes
        )

        showinfo(title="Selected File", message=self.filename)

        self.ent_file.insert(0, filename)
        self.data = np.array(pd.read_csv(self.ent_file.get(), sep=";"))
        self.stds = self.data.std(axis=1)
        self.alpha, self.loc, self.beta = gamma.fit(self.stds, floc=0)
        self.ent_alpha.configure(state="normal")
        self.ent_beta.configure(state="normal")
        self.ent_alpha.delete(0, "end")
        self.ent_beta.delete(0, "end")
        self.ent_alpha.insert(0, self.alpha)
        self.ent_beta.insert(0, self.beta)
        self.ent_alpha.configure(state="disabled")
        self.ent_beta.configure(state="disabled")

    def checkbox1(self):
        if self.var_radio.get() == 1:
            self.btn_open.configure(state="disabled")
            self.ent_alpha.configure(state="normal")
            self.ent_beta.configure(state="normal")
            self.ent_file.configure(state="disabled")
            self.ent_mean.configure(state="disabled")
            self.ent_std.configure(state="disabled")
        elif self.var_radio.get() == 2:
            self.ent_file.configure(state="normal")
            self.btn_open.configure(state="normal")
            self.ent_alpha.configure(state="disabled")
            self.ent_beta.configure(state="disabled")
            self.ent_mean.configure(state="disabled")
            self.ent_std.configure(state="disabled")
        elif self.var_radio.get() == 3:
            self.ent_mean.configure(state="normal")
            self.ent_std.configure(state="normal")
            self.ent_alpha.configure(state="disabled")
            self.ent_beta.configure(state="disabled")
            self.btn_open.configure(state="disabled")
            self.ent_file.configure(state="disabled")

    # assign infobutton commands

    def info1(self):
        self.lbl_info[
            "text"
        ] = "Insert the path of your prior patient data in here. \nThis is only needed, if the checkbox for prior data is marked. \nIf not, one can directly insert the hyperparameters below. \nThe file with the prior data must be of the shape n x k,\nwhere each new patient n is on a row and each fraction for patient n is in column k"

    def info2(self):
        self.lbl_info[
            "text"
        ] = "Insert the mean of the sparing factor distribution. \nwith this option the distribution is not updated"

    def info3(self):
        self.lbl_info[
            "text"
        ] = "Insert the standard deviation of the sparing factor distribution. \nwith this option the distribution is not updated"

    def info4(self):
        self.lbl_info[
            "text"
        ] = "Insert the shape parameter for the inverse-gamme distribution."

    def info5(self):
        self.lbl_info[
            "text"
        ] = "Insert the scale parameter for the inverse-gamme distribution."

    def infofrac(self):
        self.lbl_info[
            "text"
        ] = "Insert the number of fractions to be delivered to the patient. \n5 fractions is set a standard SBRT treatment."

    def infomin(self):
        self.lbl_info[
            "text"
        ] = "Insert the minimal physical dose that shall be delivered to the PTV95 in one fraction."

    def infomax(self):
        self.lbl_info[
            "text"
        ] = "Insert the maximal physical dose that shall be delivered to the PTV95 in one fraction."

    # assign infobutton commands

    def info10(self):
        self.lbl_info[
            "text"
        ] = "Insert the sparing factors that were observed so far.\n The sparing factor of the planning session must be included!.\nThe sparing factors must be separated by spaces e.g.:\n1.1 0.95 0.88\nFor a whole plan 6 sparing factors are needed."

    def info11(self):
        self.lbl_info["text"] = "Insert the alpha-beta ratio of the tumor tissue."

    def info12(self):
        self.lbl_info[
            "text"
        ] = "Insert the alpha-beta ratio of the dose-limiting Organ at risk."

    def info13(self):
        self.lbl_info[
            "text"
        ] = "Insert the maximum dose delivered to the dose-limiting OAR in BED."

    def info14(self):
        self.lbl_info[
            "text"
        ] = "Insert the prescribed biological effectiv dose to be delivered to the tumor."

    def info15(self):
        self.lbl_info[
            "text"
        ] = "Insert the accumulated BED so far. (If fraction one, it is zero)."

    def checkbox(self):
        if self.var.get() == 0:
            self.ent_BED.configure(state="disabled")
        else:
            self.ent_BED.configure(state="normal")

    def OAR_enable(self):
        if self.var_OAR.get() == 0:
            self.ent_OARlimit.configure(state="normal")
            self.ent_tumorlimit.configure(state="disabled")
        else:
            self.ent_OARlimit.configure(state="disabled")
            self.ent_tumorlimit.configure(state="normal")

    def compute_plan(self):
        number_of_fractions = int(self.ent_fractions.get())
        alpha = float(self.ent_alpha.get())
        beta = float(self.ent_beta.get())
        min_dose = float(self.ent_mindose.get())
        max_dose = float(self.ent_maxdose.get())
        try:
            global lbl_output
            self.lbl_output.destroy()
        except:
            pass
        if self.var_radio.get() != 3:
            fixed_prob = 0
            fixed_mean = 0
            fixed_std = 0
        if self.var_radio.get() == 3:
            fixed_prob = 1
            fixed_mean = float(self.ent_mean.get())
            fixed_std = float(self.ent_std.get())
        if self.var.get() == 0:
            try:
                alpha = float(self.ent_alpha.get())
                beta = float(self.ent_beta.get())
                sparing_factors_str = (self.ent_sf.get()).split()
                sparing_factors = [float(i) for i in sparing_factors_str]
                abt = float(self.ent_abt.get())
                abn = float(self.ent_abn.get())
                OAR_limit = float(self.ent_OARlimit.get())
                if self.var_OAR.get() == 0:
                    [tumor_doses, OAR_doses, physical_doses] = inttumor.whole_plan(
                        number_of_fractions,
                        sparing_factors,
                        abt,
                        abn,
                        alpha,
                        beta,
                        OAR_limit,
                        min_dose,
                        max_dose,
                        fixed_prob,
                        fixed_mean,
                        fixed_std,
                    )

                    self.lbl_output = tk.Frame(master=self.frame.interior)
                    self.lbl_output.pack()
                    frame = tk.Frame(
                        master=self.lbl_output, relief=tk.RAISED, borderwidth=1
                    )
                    frame.grid(row=0, column=0)
                    label = tk.Label(master=frame, text="fraction number")
                    label.pack()
                    frame = tk.Frame(
                        master=self.lbl_output, relief=tk.RAISED, borderwidth=1
                    )
                    frame.grid(row=0, column=1)
                    label = tk.Label(master=frame, text="sparing factor")
                    label.pack()
                    frame = tk.Frame(
                        master=self.lbl_output, relief=tk.RAISED, borderwidth=1
                    )
                    frame.grid(row=0, column=2)
                    label = tk.Label(
                        master=frame, text="physical dose delivered to PTV95"
                    )
                    label.pack()
                    frame = tk.Frame(
                        master=self.lbl_output, relief=tk.RAISED, borderwidth=1
                    )
                    frame.grid(row=0, column=3)
                    label = tk.Label(master=frame, text="BED delivered to tumor")
                    label.pack()
                    frame = tk.Frame(
                        master=self.lbl_output, relief=tk.RAISED, borderwidth=1
                    )
                    frame.grid(row=0, column=4)
                    label = tk.Label(master=frame, text="BED delivered to OAR")
                    label.pack()
                    for i in range(1, number_of_fractions + 1):
                        for j in range(5):
                            if j == 0:
                                frame = tk.Frame(master=self.lbl_output)
                                frame.grid(row=i, column=0)
                                label = tk.Label(master=frame, text=f"fraction {i}")
                                label.pack()
                            elif j == 1:
                                frame = tk.Frame(master=self.lbl_output)
                                frame.grid(row=i, column=1)
                                label = tk.Label(
                                    master=frame, text=f" {sparing_factors[i]}"
                                )
                                label.pack()
                            elif j == 2:
                                frame = tk.Frame(master=self.lbl_output)
                                frame.grid(row=i, column=2)
                                label = tk.Label(
                                    master=frame,
                                    text=f" {np.round(physical_doses[i-1],2)}",
                                )
                                label.pack()
                            elif j == 3:
                                frame = tk.Frame(master=self.lbl_output)
                                frame.grid(row=i, column=3)
                                label = tk.Label(
                                    master=frame,
                                    text=f" {np.round(tumor_doses[i-1],2)}",
                                )
                                label.pack()
                            elif j == 4:
                                frame = tk.Frame(master=self.lbl_output)
                                frame.grid(row=i, column=4)
                                label = tk.Label(
                                    master=frame, text=f" {np.round(OAR_doses[i-1],2)}"
                                )
                                label.pack()
                    frame = tk.Frame(
                        master=self.lbl_output, relief=tk.RAISED, borderwidth=1
                    )
                    frame.grid(row=number_of_fractions + 1, column=0)
                    label = tk.Label(master=frame, text="accumulated doses")
                    label.pack()
                    frame = tk.Frame(
                        master=self.lbl_output, relief=tk.RAISED, borderwidth=1
                    )
                    frame.grid(row=number_of_fractions + 1, column=3)
                    label = tk.Label(
                        master=frame, text=f" {np.round(np.sum(tumor_doses),2)}"
                    )
                    label.pack()
                    frame = tk.Frame(
                        master=self.lbl_output, relief=tk.RAISED, borderwidth=1
                    )
                    frame.grid(row=number_of_fractions + 1, column=4)
                    label = tk.Label(master=frame, text=f" {np.sum(OAR_doses)}")
                    label.pack()
                elif self.var_OAR.get() == 1:
                    prescribed_tumor_dose = float(self.ent_tumorlimit.get())
                    [tumor_doses, OAR_doses, physical_doses] = intOAR.whole_plan(
                        number_of_fractions,
                        sparing_factors,
                        alpha,
                        beta,
                        prescribed_tumor_dose,
                        abt,
                        abn,
                        min_dose,
                        max_dose,
                        fixed_prob,
                        fixed_mean,
                        fixed_std,
                    )
                    self.lbl_output = tk.Frame(master=self.frame.interior)
                    self.lbl_output.pack()
                    frame = tk.Frame(
                        master=self.lbl_output, relief=tk.RAISED, borderwidth=1
                    )
                    frame.grid(row=0, column=0)
                    label = tk.Label(master=frame, text="fraction number")
                    label.pack()
                    frame = tk.Frame(
                        master=self.lbl_output, relief=tk.RAISED, borderwidth=1
                    )
                    frame.grid(row=0, column=1)
                    label = tk.Label(master=frame, text="sparing factor")
                    label.pack()
                    frame = tk.Frame(
                        master=self.lbl_output, relief=tk.RAISED, borderwidth=1
                    )
                    frame.grid(row=0, column=2)
                    label = tk.Label(
                        master=frame, text="physical dose delivered to PTV95"
                    )
                    label.pack()
                    frame = tk.Frame(
                        master=self.lbl_output, relief=tk.RAISED, borderwidth=1
                    )
                    frame.grid(row=0, column=3)
                    label = tk.Label(master=frame, text="BED delivered to tumor")
                    label.pack()
                    frame = tk.Frame(
                        master=self.lbl_output, relief=tk.RAISED, borderwidth=1
                    )
                    frame.grid(row=0, column=4)
                    label = tk.Label(master=frame, text="BED delivered to OAR")
                    label.pack()
                    for i in range(1, number_of_fractions + 1):
                        for j in range(5):
                            if j == 0:
                                frame = tk.Frame(master=self.lbl_output)
                                frame.grid(row=i, column=0)
                                label = tk.Label(master=frame, text=f"fraction {i}")
                                label.pack()
                            elif j == 1:
                                frame = tk.Frame(master=self.lbl_output)
                                frame.grid(row=i, column=1)
                                label = tk.Label(
                                    master=frame, text=f" {sparing_factors[i]}"
                                )
                                label.pack()
                            elif j == 2:
                                frame = tk.Frame(master=self.lbl_output)
                                frame.grid(row=i, column=2)
                                label = tk.Label(
                                    master=frame,
                                    text=f" {np.round(physical_doses[i-1],2)}",
                                )
                                label.pack()
                            elif j == 3:
                                frame = tk.Frame(master=self.lbl_output)
                                frame.grid(row=i, column=3)
                                label = tk.Label(
                                    master=frame,
                                    text=f" {np.round(tumor_doses[i-1],2)}",
                                )
                                label.pack()
                            elif j == 4:
                                frame = tk.Frame(master=self.lbl_output)
                                frame.grid(row=i, column=4)
                                label = tk.Label(
                                    master=frame, text=f" {np.round(OAR_doses[i-1],2)}"
                                )
                                label.pack()
                                frame = tk.Frame(
                                    master=self.lbl_output,
                                    relief=tk.RAISED,
                                    borderwidth=1,
                                )
                    frame.grid(row=number_of_fractions + 1, column=0)
                    label = tk.Label(master=frame, text="accumulated doses")
                    label.pack()
                    frame = tk.Frame(
                        master=self.lbl_output, relief=tk.RAISED, borderwidth=1
                    )
                    frame.grid(row=number_of_fractions + 1, column=3)
                    label = tk.Label(
                        master=frame, text=f" {np.round(np.sum(tumor_doses),2)}"
                    )
                    label.pack()
                    frame = tk.Frame(
                        master=self.lbl_output, relief=tk.RAISED, borderwidth=1
                    )
                    frame.grid(row=number_of_fractions + 1, column=4)
                    label = tk.Label(
                        master=frame, text=f" {np.round(np.sum(OAR_doses),1)}"
                    )
                    label.pack()
            except ValueError:
                self.lbl_info[
                    "text"
                ] = "please enter correct values\nsparing factors have to been inserted with space in between. No additional brackets needed."
        else:
            try:
                sparing_factors_str = (self.ent_sf.get()).split()
                sparing_factors = [float(i) for i in sparing_factors_str]
                abt = float(self.ent_abt.get())
                abn = float(self.ent_abn.get())
                OAR_limit = float(self.ent_OARlimit.get())
                BED = float(self.ent_BED.get())
                if self.var_OAR.get() == 0:
                    [
                        Values,
                        policy,
                        actual_value,
                        actual_policy,
                        dose_delivered_OAR,
                        dose_delivered_tumor,
                        total_dose_delivered_OAR,
                        actual_dose_delivered,
                    ] = inttumor.value_eval(
                        len(sparing_factors) - 1,
                        number_of_fractions,
                        BED,
                        sparing_factors,
                        alpha,
                        beta,
                        abt,
                        abn,
                        OAR_limit,
                        min_dose,
                        max_dose,
                        fixed_prob,
                        fixed_mean,
                        fixed_std,
                    )
                    self.lbl_info[
                        "text"
                    ] = f"Optimal dose for fraction {len(sparing_factors)-1} = {actual_dose_delivered}\ndelivered tumor BED = {dose_delivered_tumor}\ndelivered OAR BED = {dose_delivered_OAR}"
                elif self.var_OAR.get() == 1:
                    prescribed_tumor_dose = float(self.ent_tumorlimit.get())
                    [
                        policy,
                        sf,
                        physical_dose,
                        tumor_dose,
                        OAR_dose,
                    ] = intOAR.value_eval(
                        len(sparing_factors) - 1,
                        number_of_fractions,
                        BED,
                        sparing_factors,
                        alpha,
                        beta,
                        prescribed_tumor_dose,
                        abt,
                        abn,
                        min_dose,
                        max_dose,
                        fixed_prob,
                        fixed_mean,
                        fixed_std,
                    )
                    self.lbl_info[
                        "text"
                    ] = f"The optimal dose for fraction {len(sparing_factors)-1},  = {physical_dose}\n dose delivered to tumor = {tumor_dose}\n dose delivered OAR = {OAR_dose}"
            except ValueError:
                self.lbl_info[
                    "text"
                ] = "please enter correct values\nsparing factors have to been inserted with space in between. No additional brackets needed."


if __name__ == "__main__":
    root = tk.Tk()
    GUI = GUI2Dextended(root)
    # Start the application
    root.mainloop()



# -*- coding: utf-8 -*-
"""
This function computes the doses for a whole
treatment plan retrospectively 
(when all sparing factors are known)

or for a single fraction with known previous
sparing factors and accumulated BED
"""

import numpy as np
import adaptfx as afx
nme = __name__

def multiple(algorithm, keys, sets=afx.SETTING_DICT):
    if isinstance(keys, dict):
        # check if keys is a dictionary from manual user
        keys = afx.DotDict(keys)

    if isinstance(sets, dict):
        # check if keys is a dictionary from manual user
        sets = afx.DotDict(sets)

    if keys.fraction != 0:
        # if only a specific fraction should be calculated
        fractions_list = np.array([keys.fraction])
        physical_doses = np.zeros(1)
        tumor_doses = np.zeros(1)
        oar_doses = np.zeros(1)
    else:
        # for calculation whole treatment in retrospect
        fractions_list = np.arange(1, keys.number_of_fractions + 1, 1)
        physical_doses = np.zeros(keys.number_of_fractions)
        tumor_doses = np.zeros(keys.number_of_fractions)
        oar_doses = np.zeros(keys.number_of_fractions)

    first_tumor_dose = keys.accumulated_tumor_dose
    first_oar_dose = keys.accumulated_oar_dose

    output_whole = afx.DotDict({})

    for i, keys.fraction in enumerate(fractions_list):
        keys.sparing_factors_public = keys.sparing_factors[0 : keys.fraction + 1]
        if algorithm == 'oar':
            output = afx.min_oar_bed(keys, sets)
        elif algorithm == 'frac':
            output = afx.min_n_frac(keys, sets)
            
        elif algorithm == 'oar_old':
            output = afx.min_oar_bed_old(keys, sets)
        elif algorithm == 'frac_old':
            output = afx.min_n_frac_old(keys, sets)
        elif algorithm == 'tumor_old':
            output = afx.max_tumor_bed_old(keys, sets)
        elif algorithm == 'tumor_oar_old':
            output = afx.min_oar_max_tumor_old(keys, sets)
        else:
            afx.aft_error('no valid algorithm given', nme)

        physical_doses[i] = output.physical_dose
        tumor_doses[i] = output.tumor_dose
        oar_doses[i] = output.oar_dose

        keys.accumulated_tumor_dose = np.nansum(tumor_doses) + first_tumor_dose
        keys.accumulated_oar_dose = np.nansum(oar_doses) + first_oar_dose

        # user specifies to plot policy number, if equal to fraction plot
        # if both zero than the user doesn't want to plot policy
        if sets.plot_policy == keys.fraction:
            output_whole.policy = output.policy
            output_whole.policy.fractions = fractions_list[sets.plot_policy - 1:]
        if sets.plot_values == keys.fraction:
            output_whole.value = output.value
            output_whole.value.fractions = fractions_list[sets.plot_values - 1:]
        if sets.plot_remains == keys.fraction:
            output_whole.remains = output.remains
            output_whole.remains.fractions = fractions_list[sets.plot_remains - 1:]

    # store doses
    exponent = afx.find_exponent(sets.dose_stepsize) - 1
    [output_whole.physical_doses, output_whole.tumor_doses, output_whole.oar_doses] = np.around(
        [physical_doses, tumor_doses, oar_doses], -exponent)
    output_whole.oar_sum, output_whole.tumor_sum = np.around(
        [np.nansum(oar_doses), np.nansum(tumor_doses)], -exponent)
    output_whole.fractions_used = np.count_nonzero(~np.isnan(output_whole.physical_doses))

    return output_whole